import logging
from typing import Any, List, Optional, Union
import pytest
from cfn_docgen.adapters.cfn_specification_repository import CfnSpecificationRepository
from cfn_docgen.adapters.internal.cache import LocalFileCache
from cfn_docgen.adapters.internal.file_loader import RemoteFileLoader
from cfn_docgen.config import AppConfig, AppContext, AwsConnectionSettings, ConnectionSettings
from cfn_docgen.domain.model.cfn_document_generator import CfnMarkdownDocumentGenerator, PropertyField

from cfn_docgen.domain.model.cfn_template import CfnTemplateConditionsNode, CfnTemplateDefinition, CfnTemplateMappingsNode, CfnTemplateMetadataCfnDocgenDefinition, CfnTemplateMetadataDefinition, CfnTemplateMetadataInterface, CfnTemplateMetadataParameterGroup, CfnTemplateMetadataParameterGroupLabel, CfnTemplateOutputDefinition, CfnTemplateOutputExportDefinition, CfnTemplateOutputsNode, CfnTemplateParameterDefinition, CfnTemplateParametersNode, CfnTemplateResourceDefinition, CfnTemplateResourceMetadataCfnDocgenDefinition, CfnTemplateResourceMetadataDefinition, CfnTemplateResourcesNode, CfnTemplateRuleAssertDefinition, CfnTemplateRuleDefinition, CfnTemplateRulesNode, CfnTemplateSource, CfnTemplateTree

@pytest.fixture
def context():
    return AppContext(
        log_level=logging.DEBUG,
        connection_settings=ConnectionSettings(aws=AwsConnectionSettings(profile_name=None)),
    )

@pytest.fixture
def cfn_template_tree(context:AppContext):
    spec_repository = CfnSpecificationRepository(
        source_url=AppConfig.DEFAULT_SPECIFICATION_URL,
        loader=RemoteFileLoader(context=context),
        cache=LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=context),
        recursive_resource_types=AppConfig.RECURSIVE_RESOURCE_TYPES,
        context=context,
    )

    definition = CfnTemplateDefinition(
        AWSTemplateFormatVersion="2012-10",
        Description="template-description",
        Metadata=CfnTemplateMetadataDefinition(**{
            'AWS::CloudFormation::Interface': CfnTemplateMetadataInterface(
                ParameterGroups=[
                    CfnTemplateMetadataParameterGroup(
                        Label=CfnTemplateMetadataParameterGroupLabel(
                            default="Group1"
                        ),
                        Parameters=["Param1"]
                    )
                ]
            ),
            "CfnDocgen": CfnTemplateMetadataCfnDocgenDefinition(
                Description="- line1\n- line2",
                Mappings={
                    "Map1": "map1-description"
                },
                Rules={
                    "Rule1": "rule1-description"
                },
                Conditions={
                    "Cond1": "cond1-description"
                },
            )
        }), # type: ignore
        Parameters={
            "Param1": CfnTemplateParameterDefinition(
                Type="String"
            ),
            "Param2": CfnTemplateParameterDefinition(
                Type="Number"
            )
        },
        Mappings={
            "Map1": {
                "name1": {
                    "key1": "value1"
                }
            }
        },
        Rules={
            "Rule1": CfnTemplateRuleDefinition(
                Assertions=[
                    CfnTemplateRuleAssertDefinition(
                        Assert={
                            "assert1": "value1"
                        },
                        AssertDescription="assert1-description"
                    )
                ]
            )
        },
        Conditions={
            "Cond1": {
                "key1": "value1"
            }
        },
        Resources={
            "Resource1": CfnTemplateResourceDefinition(
                Metadata=CfnTemplateResourceMetadataDefinition(
                    CfnDocgen=CfnTemplateResourceMetadataCfnDocgenDefinition(
                        Description="resource1-description",
                        Properties={
                            "BlockDeviceMappings": [
                                {
                                    "Ebs": {
                                        "Encrypted": "encrypted-description"
                                    }
                                }
                            ]
                        },
                    )
                ),
                Type="AWS::EC2::Instance",
                Properties={
                    "BlockDeviceMappings": [
                        {
                            "Ebs": {
                                "Encrypted": True
                            }
                        }
                    ],
                    "ImageId": {"Ref": "param1"}
                }
            ),
            "Resource2": CfnTemplateResourceDefinition(
                Type="AWS::CertificateManager::Certificate",
                Properties={}
            )
        },
        Outputs={
            "Out1": CfnTemplateOutputDefinition(
                Value={"Ref": "resource1"}
            )
        }
    )

    return CfnTemplateTree(
        template_source=CfnTemplateSource("/tmp/template.yaml", context=AppContext(log_level=logging.DEBUG)),
        definition=definition,
        spec_repository=spec_repository,
        context=context
    )

def test_CfnMarkdownDocumentGenerator_header(
    context:AppContext,
):
    generator = CfnMarkdownDocumentGenerator(context=context)
    header = generator.header()
    from cfn_docgen import __version__
    expected = f"<!-- cfn-docgen: generated by v{__version__} -->"
    assert header == expected

@pytest.mark.parametrize("original,expected", [
    ("template.yaml", "templateyaml"),
    ("Map_1", "map_1"),
    ("Resource (AWS::EC2::Instance)", "resource-awsec2instance"),
])
def test_CfnMarkdownDocumentGenerator_toc_escape(
    original:str, expected:str,
    context:AppContext,
):
    generator = CfnMarkdownDocumentGenerator(context=context)
    escape = generator._toc_escape(original) # type: ignore
    assert escape == expected 


def test_CfnMarkdownDocumentGenerator_table_of_contents(
    cfn_template_tree:CfnTemplateTree,
    context:AppContext,
):

    generator = CfnMarkdownDocumentGenerator(context=context)
    expected_toc = "\n".join([
        "- [template.yaml](#templateyaml)",
        "  - [Description](#description)",
        "  - [Parameters](#parameters)",
        "    - [Group1](#group1)",
        "      - [Param1](#param1)",
        "    - [Param2](#param2)",
        "  - [Mappings](#mappings)",
        "    - [Map1](#map1)",
        "  - [Conditions](#conditions)",
        "    - [Cond1](#cond1)",
        "  - [Rules](#rules)",
        "    - [Rule1](#rule1)",
        "  - [Resources](#resources)",
        "    - [Resource2 (AWS::CertificateManager::Certificate)](#resource2-awscertificatemanagercertificate)",
        "    - [Resource1 (AWS::EC2::Instance)](#resource1-awsec2instance)",
        "  - [Outputs](#outputs)",
        "    - [Out1](#out1)"
    ])
    toc = generator.table_of_contents(cfn_template_tree)
    assert toc == expected_toc


def test_CfnMarkdownDocumentGenerator_table_of_contents_with_resource_groups(
    context:AppContext,
):

    generator = CfnMarkdownDocumentGenerator(context=context)
    expected_toc = "\n".join([
        "- [template.yaml](#templateyaml)",
        "  - [Description](#description)",
        "  - [Parameters](#parameters)",
        "  - [Mappings](#mappings)",
        "  - [Conditions](#conditions)",
        "  - [Rules](#rules)",
        "  - [Resources](#resources)",
        "    - [ResourceGroup1](#resourcegroup1)",
        "      - [Resource12 (AWS::CertificateManager::Certificate)](#resource12-awscertificatemanagercertificate)",
        "      - [Resource11 (AWS::EC2::Instance)](#resource11-awsec2instance)",
        "    - [ResourceGroup2](#resourcegroup2)",
        "      - [Resource22 (AWS::CertificateManager::Certificate)](#resource22-awscertificatemanagercertificate)",
        "      - [Resource21 (AWS::EC2::Instance)](#resource21-awsec2instance)",
        "  - [Outputs](#outputs)",
    ])
    spec_repository = CfnSpecificationRepository(
        source_url=AppConfig.DEFAULT_SPECIFICATION_URL,
        loader=RemoteFileLoader(context=context),
        cache=LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=context),
        recursive_resource_types=AppConfig.RECURSIVE_RESOURCE_TYPES,
        context=context,
    )

    definition = CfnTemplateDefinition(
        Resources={
            "Resource11": CfnTemplateResourceDefinition(
                Metadata=CfnTemplateResourceMetadataDefinition(**{
                    "aws:cdk:path": "some-stack/ResourceGroup1/resource11"
                }), # type: ignore
                Type="AWS::EC2::Instance",
                Properties={}
            ),
            "Resource12": CfnTemplateResourceDefinition(
                Metadata=CfnTemplateResourceMetadataDefinition(**{
                    "aws:cdk:path": "some-stack/ResourceGroup1/resource12"
                }), # type: ignore
                Type="AWS::CertificateManager::Certificate",
                Properties={}
            ),
            "Resource21": CfnTemplateResourceDefinition(
                Metadata=CfnTemplateResourceMetadataDefinition(**{
                    "aws:cdk:path": "some-stack/ResourceGroup2/resource21"
                }), # type: ignore
                Type="AWS::EC2::Instance",
                Properties={}
            ),
            "Resource22": CfnTemplateResourceDefinition(
                Metadata=CfnTemplateResourceMetadataDefinition(**{
                    "aws:cdk:path": "some-stack/ResourceGroup2/resource22"
                }), # type: ignore
                Type="AWS::CertificateManager::Certificate",
                Properties={}
            )
        },
    )
    tree = CfnTemplateTree(
        template_source=CfnTemplateSource("/tmp/template.yaml", context=AppContext(log_level=logging.DEBUG)),
        definition=definition,
        spec_repository=spec_repository,
        context=context
    )
    toc = generator.table_of_contents(tree)
    assert toc == expected_toc

@pytest.mark.parametrize("version,description,transform,e_version,e_description,e_transform", [
    ("2012-10", "template-descirption", "t1", "2012-10", "template-descirption", "t1"),
    ("2012-10", "template-descirption", ["t1", "t2"], "2012-10", "template-descirption", "<ul><li>t1</li><li>t2</li></ul>"),
    (None, None, [], "-", "-", "-"),
])
def test_CfnMarkdownDocumentGenerator_overview(
    version:Optional[str], description:Optional[str], transform:Union[str, List[str]],
    e_version:str, e_description:str, e_transform:str,
    cfn_template_tree:CfnTemplateTree,
    context:AppContext,
):
    generator = CfnMarkdownDocumentGenerator(context=context)
    expected_overview = "\n".join([
        "# template.yaml",
        "",
        "| | |",
        "|-|-|",
        f"|AWSTemplateFormatVersion|{e_version}|",
        f"|Description|{e_description}|",
        f"|Transform|{e_transform}|",
    ])
    cfn_template_tree.aws_template_format_version = version
    cfn_template_tree.description = description
    cfn_template_tree.transform = transform
    overview = generator.overview(cfn_template_tree)
    assert overview == expected_overview

@pytest.mark.parametrize("cfn_docgen_description,e_cfn_docgen_description",[
    (
        "cfn-docgen-description",
        "\n".join([
            "## Description",
            "",
            "cfn-docgen-description",
        ]),
    ),
    (
        None,
        "\n".join([
            "## Description",
            "",
        ]),
    ),
    (
        "- d1\n- d2",
        "\n".join([
            "## Description",
            "",
            "- d1\n- d2",
        ]),
    )
])
def test_CfnMarkdownDocumentGenerator_description(
        cfn_docgen_description:Optional[str], e_cfn_docgen_description:str,
        cfn_template_tree:CfnTemplateTree,
        context:AppContext,
):
    generator = CfnMarkdownDocumentGenerator(context=context)
    cfn_template_tree.cfn_docgen_description = cfn_docgen_description
    description = generator.description(cfn_template_tree)
    assert description == e_cfn_docgen_description

@pytest.mark.parametrize("paramters_node,e_parameters",[
    (
        # without parameter groups
        CfnTemplateParametersNode(
            context=AppContext(log_level=logging.DEBUG),
            definitions={
                "param1": CfnTemplateParameterDefinition(
                    AllowedPattern="allowedpattern",
                    AllowedValues=["s1", "s2"],
                    ConstraintDescription="constraintdescription",
                    Default="s1",
                    Description="param1-description",
                    MaxLength=10,
                    MinLength=2,
                    MaxValue=None,
                    MinValue=None,
                    NoEcho=True,
                    Type="String",
                ),
                "param2": CfnTemplateParameterDefinition(Type="Number"),
            },
            parameter_groups=[]
        ),
        "\n".join([
            "## Parameters",
            "",
            "### param1",
            "",
            "param1-description",
            "",
            "|Type|Default|AllowedValues|AllowedPattern|NoEcho|MinValue|MaxValue|MinLength|MaxLength|ConstraintDescription|",
            "|-|-|-|-|-|-|-|-|-|-|",
            "|String|s1|<ul><li>s1</li><li>s2</li></ul>|allowedpattern|true|-|-|2|10|constraintdescription|",
            "",
            "### param2",
            "",
            "",
            "",
            "|Type|Default|AllowedValues|AllowedPattern|NoEcho|MinValue|MaxValue|MinLength|MaxLength|ConstraintDescription|",
            "|-|-|-|-|-|-|-|-|-|-|",
            "|Number|-|-|-|false|-|-|-|-|-|",
        ])
    ),
    (
        # with parameter groups
        CfnTemplateParametersNode(
            context=AppContext(log_level=logging.DEBUG),
            definitions={
                "param1": CfnTemplateParameterDefinition(
                    AllowedPattern="allowedpattern",
                    AllowedValues=["s1", "s2"],
                    ConstraintDescription="constraintdescription",
                    Default="s1",
                    Description="param1-description",
                    MaxLength=10,
                    MinLength=2,
                    MaxValue=None,
                    MinValue=None,
                    NoEcho=True,
                    Type="String",
                ),
                "param2": CfnTemplateParameterDefinition(Type="Number"),
            },
            parameter_groups=[
                CfnTemplateMetadataParameterGroup(
                    Label=CfnTemplateMetadataParameterGroupLabel(default="Group1"),
                    Parameters=["param1", "param2"]
                )
            ]
        ),
        "\n".join([
            "## Parameters",
            "",
            "### Group1",
            "",
            "#### param1",
            "",
            "param1-description",
            "",
            "|Type|Default|AllowedValues|AllowedPattern|NoEcho|MinValue|MaxValue|MinLength|MaxLength|ConstraintDescription|",
            "|-|-|-|-|-|-|-|-|-|-|",
            "|String|s1|<ul><li>s1</li><li>s2</li></ul>|allowedpattern|true|-|-|2|10|constraintdescription|",
            "",
            "#### param2",
            "",
            "",
            "",
            "|Type|Default|AllowedValues|AllowedPattern|NoEcho|MinValue|MaxValue|MinLength|MaxLength|ConstraintDescription|",
            "|-|-|-|-|-|-|-|-|-|-|",
            "|Number|-|-|-|false|-|-|-|-|-|",
        ])
    ),
    (
        # partially with parameter groups
        CfnTemplateParametersNode(
            context=AppContext(log_level=logging.DEBUG),
            definitions={
                "param1": CfnTemplateParameterDefinition(
                    AllowedPattern="allowedpattern",
                    AllowedValues=["s1", "s2"],
                    ConstraintDescription="constraintdescription",
                    Default="s1",
                    Description="param1-description",
                    MaxLength=10,
                    MinLength=2,
                    MaxValue=None,
                    MinValue=None,
                    NoEcho=True,
                    Type="String",
                ),
                "param2": CfnTemplateParameterDefinition(Type="Number"),
            },
            parameter_groups=[
                CfnTemplateMetadataParameterGroup(
                    Label=CfnTemplateMetadataParameterGroupLabel(default="Group1"),
                    Parameters=["param2"]
                )
            ]
        ),
        "\n".join([
            "## Parameters",
            "",
            "### Group1",
            "",
            "#### param2",
            "",
            "",
            "",
            "|Type|Default|AllowedValues|AllowedPattern|NoEcho|MinValue|MaxValue|MinLength|MaxLength|ConstraintDescription|",
            "|-|-|-|-|-|-|-|-|-|-|",
            "|Number|-|-|-|false|-|-|-|-|-|",
            "",
            "### param1",
            "",
            "param1-description",
            "",
            "|Type|Default|AllowedValues|AllowedPattern|NoEcho|MinValue|MaxValue|MinLength|MaxLength|ConstraintDescription|",
            "|-|-|-|-|-|-|-|-|-|-|",
            "|String|s1|<ul><li>s1</li><li>s2</li></ul>|allowedpattern|true|-|-|2|10|constraintdescription|",
        ])
    ),
    (
        # no parameters
        CfnTemplateParametersNode(
            context=AppContext(log_level=logging.DEBUG),
            definitions={},
            parameter_groups=[]
        ),
        "\n".join([
            "## Parameters",
        ])
    )
])
def test_CfnMarkdownDocumentGenerator_parameters(
    paramters_node:CfnTemplateParametersNode, e_parameters:str,
    cfn_template_tree:CfnTemplateTree,
    context:AppContext,
):
    generator = CfnMarkdownDocumentGenerator(context=context)
    cfn_template_tree.parameters_node = paramters_node

    parameters = generator.parameters(cfn_template_tree)
    assert parameters == e_parameters

@pytest.mark.parametrize("mappings_node,e_mappings", [
    (
        CfnTemplateMappingsNode(
            context=AppContext(log_level=logging.DEBUG),
            definitions={
                "map1": {
                    "name1": {
                        "key11": "value11",
                        "key12": "value12",
                    }
                },
                "map2": {
                    "name2": {
                        "key2": "value2"
                    }
                }
            },
            descriptions={
                "map1": "map1-description"
            }
        ),
        "\n".join([
            "## Mappings",
            "",
            "### map1",
            "",
            "map1-description",
            "",
            "|Map|Key|Value|",
            "|-|-|-|",
            "|name1|key11|value11|",
            "|name1|key12|value12|",
            "",
            "### map2",
            "",
            "",
            "",
            "|Map|Key|Value|",
            "|-|-|-|",
            "|name2|key2|value2|",
        ])
    ),
    (
        CfnTemplateMappingsNode(
            context=AppContext(log_level=logging.DEBUG),
            definitions={},
            descriptions={}
        ),
        "\n".join([
            "## Mappings",
        ])
    )
])
def test_CfnMarkdownDocumentGenerator_mappings(
        mappings_node:CfnTemplateMappingsNode, e_mappings:str,
        cfn_template_tree:CfnTemplateTree,
        context:AppContext,
):
    generator = CfnMarkdownDocumentGenerator(context=context)
    cfn_template_tree.mappings_node = mappings_node
    
    mappings = generator.mappings(cfn_template_tree)
    assert mappings == e_mappings


@pytest.mark.parametrize("conditions_node,e_conditions", [
    (
        CfnTemplateConditionsNode(
            context=AppContext(log_level=logging.DEBUG),
            definitions={
                "cond1": {
                    "Fn::Equals": [
                        {"Ref": "AWS::Region"},
                        "ap-northeast-1"
                    ]
                },
                "cond2": {
                    "Fn::Equals": [1, 1]
                }
            },
            descriptions={
                "cond1": "cond1-description"
            }
        ),
        "\n".join([
            "## Conditions",
            "",
            "### cond1",
            "",
            "cond1-description",
            "",
            "|Condition|",
            "|-|",
            '|{<br/>&nbsp;&nbsp;"Fn::Equals":&nbsp;[<br/>&nbsp;&nbsp;&nbsp;&nbsp;{<br/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"Ref":&nbsp;"AWS::Region"<br/>&nbsp;&nbsp;&nbsp;&nbsp;},<br/>&nbsp;&nbsp;&nbsp;&nbsp;"ap-northeast-1"<br/>&nbsp;&nbsp;]<br/>}|',
            "",
            "### cond2",
            "",
            "",
            "",
            "|Condition|",
            "|-|",
            '|{<br/>&nbsp;&nbsp;"Fn::Equals":&nbsp;[<br/>&nbsp;&nbsp;&nbsp;&nbsp;1,<br/>&nbsp;&nbsp;&nbsp;&nbsp;1<br/>&nbsp;&nbsp;]<br/>}|',
        ])
    ),
    (
        CfnTemplateConditionsNode(
            context=AppContext(log_level=logging.DEBUG),
            definitions={},
            descriptions={}
        ),
        "\n".join([
            "## Conditions",
        ])
    )
])
def test_CfnMarkdownDocumentGenerator_conditions(
        conditions_node:CfnTemplateConditionsNode, e_conditions:str,
        cfn_template_tree:CfnTemplateTree,
        context:AppContext,
):
    generator = CfnMarkdownDocumentGenerator(context=context)
    cfn_template_tree.conditions_node = conditions_node
    
    conditions = generator.conditions(cfn_template_tree)
    assert conditions == e_conditions

@pytest.mark.parametrize("j,expected", [
    (
        {
            "Fn::Equals": [
                {
                    "Ref": "AWS::Region"
                },
                "ap-northeast-1"
            ]
        },
        '{<br/>&nbsp;&nbsp;"Fn::Equals":&nbsp;[<br/>&nbsp;&nbsp;&nbsp;&nbsp;{<br/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"Ref":&nbsp;"AWS::Region"<br/>&nbsp;&nbsp;&nbsp;&nbsp;},<br/>&nbsp;&nbsp;&nbsp;&nbsp;"ap-northeast-1"<br/>&nbsp;&nbsp;]<br/>}',
    ),
    (
        [
            {
                "Ref": "AWS::NoValue"
            },
            {
                "Ref": "AWS::Region"
            }
        ],
        '[<br/>&nbsp;&nbsp;{<br/>&nbsp;&nbsp;&nbsp;&nbsp;"Ref":&nbsp;"AWS::NoValue"<br/>&nbsp;&nbsp;},<br/>&nbsp;&nbsp;{<br/>&nbsp;&nbsp;&nbsp;&nbsp;"Ref":&nbsp;"AWS::Region"<br/>&nbsp;&nbsp;}<br/>]',
    ),
    (
        "string",
        "string"
    ),
    (
        True,
        "true"
    ),
    (
        "${string}",
        "\\${string}"
    ),
    (
        {"Sub": "${string}"},
        '{<br/>&nbsp;&nbsp;"Sub":&nbsp;"\\${string}"<br/>}',
    ),
    (
        "string|string",
        "string\\|string"
    ),
    (
        {"Ref": "string|string"},
        '{<br/>&nbsp;&nbsp;"Ref":&nbsp;"string\\|string"<br/>}',
    ),
    (
        "日本語",
        "日本語"
    ),
    (
        {"Ref": "日本語"},
        '{<br/>&nbsp;&nbsp;"Ref":&nbsp;"日本語"<br/>}',
    ),
    (
        "List<AWS::EC2::VPC>",
        "List\\<AWS::EC2::VPC\\>"
    ),
    (
        {"Ref": "List<AWS::EC2::VPC>"},
        '{<br/>&nbsp;&nbsp;"Ref":&nbsp;"List\\<AWS::EC2::VPC\\>"<br/>}',
    )

])
def test_CfnMarkdownDocumentGenerator_dump_json(
    j:Any,expected:str,
    context:AppContext,
):
    generator = CfnMarkdownDocumentGenerator(context=context)

    dumped = generator._dump_json(j) # type: ignore
    assert dumped == expected
    

@pytest.mark.parametrize("rules_node,e_rules",[
    (
        CfnTemplateRulesNode(
            context=AppContext(log_level=logging.DEBUG),
            definitions={
                "rule1": CfnTemplateRuleDefinition(
                    RuleCondition={
                        "Fn::Equals": [
                            {
                                "Ref": "AWS::Region"
                            },
                            "ap-northeast-1"
                        ]
                    },
                    Assertions=[
                        CfnTemplateRuleAssertDefinition(
                            Assert={
                                "Fn::Equals": [
                                    {
                                        "Ref": "AWS::Region"
                                    },
                                    "ap-northeast-1"
                                ]
                            },
                            AssertDescription="assert-description1"
                        ),
                        CfnTemplateRuleAssertDefinition(
                            Assert={
                                "Fn::Equals": [
                                    {
                                        "Ref": "AWS::Region"
                                    },
                                    "ap-northeast-1"
                                ]
                            },
                            AssertDescription="assert-description2"
                        )
                    ]
                ),
                "rule2": CfnTemplateRuleDefinition(
                    RuleCondition=None,
                    Assertions=[
                        CfnTemplateRuleAssertDefinition(
                            Assert={"str":"str"},
                            AssertDescription="assert-description2"
                        )
                    ]
                )
            },
            descriptions={
                "rule1": "rule1-description"
            }
        ),
        "\n".join([
            "## Rules",
            "",
            "### rule1",
            "",
            "rule1-description",
            "",
            "|RuleCondition|",
            "|-|",
            '|{<br/>&nbsp;&nbsp;"Fn::Equals":&nbsp;[<br/>&nbsp;&nbsp;&nbsp;&nbsp;{<br/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"Ref":&nbsp;"AWS::Region"<br/>&nbsp;&nbsp;&nbsp;&nbsp;},<br/>&nbsp;&nbsp;&nbsp;&nbsp;"ap-northeast-1"<br/>&nbsp;&nbsp;]<br/>}|',
            "",
            "|Assert|AssertDescription|",
            "|-|-|",
            '|{<br/>&nbsp;&nbsp;"Fn::Equals":&nbsp;[<br/>&nbsp;&nbsp;&nbsp;&nbsp;{<br/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"Ref":&nbsp;"AWS::Region"<br/>&nbsp;&nbsp;&nbsp;&nbsp;},<br/>&nbsp;&nbsp;&nbsp;&nbsp;"ap-northeast-1"<br/>&nbsp;&nbsp;]<br/>}|assert-description1|',
            '|{<br/>&nbsp;&nbsp;"Fn::Equals":&nbsp;[<br/>&nbsp;&nbsp;&nbsp;&nbsp;{<br/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"Ref":&nbsp;"AWS::Region"<br/>&nbsp;&nbsp;&nbsp;&nbsp;},<br/>&nbsp;&nbsp;&nbsp;&nbsp;"ap-northeast-1"<br/>&nbsp;&nbsp;]<br/>}|assert-description2|',
            "",
            "### rule2",
            "",
            "",
            "",
            "|RuleCondition|",
            "|-|",
            "|-|",
            "",
            "|Assert|AssertDescription|",
            "|-|-|",
            '|{<br/>&nbsp;&nbsp;"str":&nbsp;"str"<br/>}|assert-description2|',
        ])
    ),
    (
        CfnTemplateRulesNode(
            context=AppContext(log_level=logging.DEBUG),
            definitions={},
            descriptions={},
        ),
        "\n".join([
            "## Rules",
        ])
    )
])
def test_CfnMarkdownDocumentGenerator_rules(
    rules_node:CfnTemplateRulesNode, e_rules:str, 
    cfn_template_tree:CfnTemplateTree,
    context:AppContext,
):
    generator = CfnMarkdownDocumentGenerator(context=context)
    cfn_template_tree.rules_node = rules_node

    rules = generator.rules(cfn_template_tree)
    assert rules == e_rules


@pytest.mark.parametrize("resources_node,e_resources", [
    (
        CfnTemplateResourcesNode(
            context=AppContext(log_level=logging.DEBUG),
            definitions={
                "Instance": CfnTemplateResourceDefinition(
                    Type="AWS::EC2::Instance",
                    DependsOn=[
                        "Vpc", "Subnet",
                    ],
                    Condition="Cond",
                    CreationPolicy={"Creation": "Policy"},
                    UpdatePolicy={"Update": "Policy"},
                    UpdateReplacePolicy="Snapshot",
                    DeletionPolicy="Retain",
                    Metadata=CfnTemplateResourceMetadataDefinition(
                        CfnDocgen=CfnTemplateResourceMetadataCfnDocgenDefinition(
                            Description="instance-description",
                            Properties={
                                "ImageId": "imageid",
                                "CpuOptions": "cpuoptions",
                                "BlockDeviceMappings": [
                                    {
                                        "Ebs": {
                                            "Encrypted": "encrypted"
                                        }
                                    }
                                ]
                            }
                        )
                    ),
                    Properties={
                        "ImageId": "IMAGEID",
                        "CpuOptions": {
                            "CoreCount": 0,
                        },
                        "BlockDeviceMappings": [
                            {
                                "Ebs": {
                                    "Encrypted": True,
                                    "VolumeType": "VOLUMETYPE",
                                }
                            },
                            {
                                "NoDevice": {}
                            }
                        ],
                        "SecurityGroupIds": [
                            "SECURITYGROUPID1",
                            "SECURITYGROUPID2",
                        ],
                        "SsmAssociations": [
                            {
                                "AssociationParameters": [
                                    {
                                        "Key": "KEY1",
                                        "Value": ["VALUE11", "VALUE12"]
                                    },
                                    {
                                        "Key": "KEY2",
                                        "Value": ["VALUE21", "VALUE22"]
                                    }
                                ]
                            }
                        ]
                    }
                ),
                "Table": CfnTemplateResourceDefinition(
                    Type="AWS::DynamoDB::Table",
                    Metadata=None,
                    Properties={
                        "TableName": "TABLENAME"
                    },
                )
            },
            spec_repository=CfnSpecificationRepository(
                context=AppContext(log_level=logging.DEBUG),
                source_url=AppConfig.DEFAULT_SPECIFICATION_URL,
                loader=RemoteFileLoader(context=AppContext(log_level=logging.DEBUG)),
                cache=LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=AppContext(log_level=logging.DEBUG)),
                recursive_resource_types=AppConfig.RECURSIVE_RESOURCE_TYPES,
            ),
            resource_groups={},
        ),
        "\n".join([
            "## Resources",
            "",
            "### [Table (AWS::DynamoDB::Table)](http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-dynamodb-table.html)",
            "",
            "",
            "",
            "|DependsOn|Condition|CreationPolicy|UpdatePolicy|UpdateReplacePolicy|DeletionPolicy|",
            "|-|-|-|-|-|-|",
            "|-|-|-|-|Delete|Delete|",
            "",
            "|Property|Value|Description|Type|Required|UpdateType|",
            "|-|-|-|-|-|-|",
            "|TableName|TABLENAME|-|String|false|Immutable|",
            "",
            "### [Instance (AWS::EC2::Instance)](http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ec2-instance.html)",
            "",
            "instance-description",
            "",
            "|DependsOn|Condition|CreationPolicy|UpdatePolicy|UpdateReplacePolicy|DeletionPolicy|",
            "|-|-|-|-|-|-|",
            '|[<br/>&nbsp;&nbsp;"Vpc",<br/>&nbsp;&nbsp;"Subnet"<br/>]|Cond|{<br/>&nbsp;&nbsp;"Creation":&nbsp;"Policy"<br/>}|{<br/>&nbsp;&nbsp;"Update":&nbsp;"Policy"<br/>}|Snapshot|Retain|',
            "",
            "|Property|Value|Description|Type|Required|UpdateType|",
            "|-|-|-|-|-|-|",
            "|BlockDeviceMappings[0]|-|-|List of BlockDeviceMapping|false|Conditional|",
            "|&nbsp;&nbsp;Ebs|-|-|Ebs|false|Mutable|",
            "|&nbsp;&nbsp;&nbsp;&nbsp;Encrypted|true|encrypted|Boolean|false|Mutable|",
            "|&nbsp;&nbsp;&nbsp;&nbsp;VolumeType|VOLUMETYPE|-|String|false|Mutable|",
            "|BlockDeviceMappings[1]|-|-|List of BlockDeviceMapping|false|Conditional|",
            "|&nbsp;&nbsp;NoDevice|-|-|NoDevice|false|Mutable|",
            "|CpuOptions|-|cpuoptions|CpuOptions|false|Immutable|",
            "|&nbsp;&nbsp;CoreCount|0|-|Integer|false|Mutable|",
            "|ImageId|IMAGEID|imageid|String|false|Immutable|",
            '|SecurityGroupIds|[<br/>&nbsp;&nbsp;"SECURITYGROUPID1",<br/>&nbsp;&nbsp;"SECURITYGROUPID2"<br/>]|-|List of String|false|Conditional|',
            '|SsmAssociations[0]|-|-|List of SsmAssociation|false|Mutable|',
            '|&nbsp;&nbsp;AssociationParameters[0]|-|-|List of AssociationParameter|false|Mutable|',
            '|&nbsp;&nbsp;&nbsp;&nbsp;Key|KEY1|-|String|true|Mutable|',
            '|&nbsp;&nbsp;&nbsp;&nbsp;Value|[<br/>&nbsp;&nbsp;"VALUE11",<br/>&nbsp;&nbsp;"VALUE12"<br/>]|-|List of String|true|Mutable|',
            '|&nbsp;&nbsp;AssociationParameters[1]|-|-|List of AssociationParameter|false|Mutable|',
            '|&nbsp;&nbsp;&nbsp;&nbsp;Key|KEY2|-|String|true|Mutable|',
            '|&nbsp;&nbsp;&nbsp;&nbsp;Value|[<br/>&nbsp;&nbsp;"VALUE21",<br/>&nbsp;&nbsp;"VALUE22"<br/>]|-|List of String|true|Mutable|',
        ])
    ),
    (
        CfnTemplateResourcesNode(
            context=AppContext(log_level=logging.DEBUG),
            definitions={},
            spec_repository=CfnSpecificationRepository(
                context=AppContext(log_level=logging.DEBUG),
                source_url=AppConfig.DEFAULT_SPECIFICATION_URL,
                loader=RemoteFileLoader(context=AppContext(log_level=logging.DEBUG)),
                cache=LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=AppContext(log_level=logging.DEBUG)),
                recursive_resource_types=AppConfig.RECURSIVE_RESOURCE_TYPES,
            ),
            resource_groups={},
        ),
        "\n".join([
            "## Resources",
        ])
    )
])
def test_CfnMarkdownDocumentGenerator_resources(
        resources_node:CfnTemplateResourcesNode, e_resources:str,
        cfn_template_tree:CfnTemplateTree,
        context:AppContext,
):
    generator = CfnMarkdownDocumentGenerator(context=context)
    cfn_template_tree.resources_node = resources_node

    resources = generator.resources(cfn_template_tree)
    assert resources == e_resources

def test_CfnMarkdownDocumentGenerator_flatten_properties(context:AppContext):
    cfn_template_tree = CfnTemplateResourcesNode(
        context=context,
        definitions={
            "Instance": CfnTemplateResourceDefinition(
                Type="AWS::EC2::Instance",
                DependsOn=[
                    "Vpc", "Subnet",
                ],
                Condition="Cond",
                CreationPolicy={"Creation": "Policy"},
                UpdatePolicy={"Update": "Policy"},
                UpdateReplacePolicy="Snapshot",
                DeletionPolicy="Retain",
                Properties={
                    "ImageId": "IMAGEID",
                    "CpuOptions": {
                        "CoreCount": 0,
                    },
                    "BlockDeviceMappings": [
                        {
                            "Ebs": {
                                "Encrypted": True,
                                "VolumeType": "VOLUMETYPE",
                            }
                        },
                        {
                            "NoDevice": {}
                        }
                    ],
                    "SecurityGroupIds": [
                        "SECURITYGROUPID1",
                        "SECURITYGROUPID2",
                    ],
                    "SsmAssociations": [
                        {
                            "AssociationParameters": [
                                {
                                    "Key": "KEY1",
                                    "Value": ["VALUE11", "VALUE12"]
                                },
                                {
                                    "Key": "KEY2",
                                    "Value": ["VALUE21", "VALUE22"]
                                }
                            ]
                        }
                    ]
                }
            )
        },
        spec_repository=CfnSpecificationRepository(
            context=context,
            source_url=AppConfig.DEFAULT_SPECIFICATION_URL,
            loader=RemoteFileLoader(context=context),
            cache=LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=context,),
            recursive_resource_types=AppConfig.RECURSIVE_RESOURCE_TYPES,
        ),
        resource_groups={},
    )
    expected:List[PropertyField] = [
        PropertyField(
            Property="BlockDeviceMappings[0]", Type="List of BlockDeviceMapping",
            Value="-", Description="-", Required="false", UpdateType="Conditional",
        ),
        PropertyField(
            Property="&nbsp;&nbsp;Ebs", Type="Ebs",
            Value="-", Description="-", Required="false", UpdateType="Mutable",
        ),
        PropertyField(
            Property="&nbsp;&nbsp;&nbsp;&nbsp;Encrypted", Type="Boolean",
            Value="true", Description="-", Required="false", UpdateType="Mutable",
        ),
        PropertyField(
            Property="&nbsp;&nbsp;&nbsp;&nbsp;VolumeType", Type="String",
            Value="VOLUMETYPE", Description="-", Required="false", UpdateType="Mutable",
        ),
        PropertyField(
            Property="BlockDeviceMappings[1]", Type="List of BlockDeviceMapping",
            Value="-", Description="-", Required="false", UpdateType="Conditional",
        ),
        PropertyField(
            Property="&nbsp;&nbsp;NoDevice", Type="NoDevice",
            Value="-", Description="-", Required="false", UpdateType="Mutable",
        ),
        PropertyField(
            Property="CpuOptions", Type="CpuOptions",
            Value="-", Description="-", Required="false", UpdateType="Immutable",
        ),
        PropertyField(
            Property="&nbsp;&nbsp;CoreCount", Type="Integer",
            Value="0", Description="-", Required="false", UpdateType="Mutable",
        ),
        PropertyField(
            Property="ImageId", Type="String",
            Value="IMAGEID", Description="-", Required="false", UpdateType="Immutable",
        ),
        PropertyField(
            Property="SecurityGroupIds", Type="List of String",
            Value='[<br/>&nbsp;&nbsp;"SECURITYGROUPID1",<br/>&nbsp;&nbsp;"SECURITYGROUPID2"<br/>]', Description="-", Required="false", UpdateType="Conditional",
        ),
        PropertyField(
            Property="SsmAssociations[0]", Type="List of SsmAssociation",
            Value="-", Description="-", Required="false", UpdateType="Mutable",
        ),
        PropertyField(
            Property="&nbsp;&nbsp;AssociationParameters[0]", Type="List of AssociationParameter",
            Value="-", Description="-", Required="false", UpdateType="Mutable",
        ),
        PropertyField(
            Property="&nbsp;&nbsp;&nbsp;&nbsp;Key", Type="String",
            Value="KEY1", Description="-", Required="true", UpdateType="Mutable",
        ),
        PropertyField(
            Property="&nbsp;&nbsp;&nbsp;&nbsp;Value", Type="List of String",
            Value='[<br/>&nbsp;&nbsp;"VALUE11",<br/>&nbsp;&nbsp;"VALUE12"<br/>]', Description="-", Required="true", UpdateType="Mutable",
        ),
        PropertyField(
            Property="&nbsp;&nbsp;AssociationParameters[1]", Type="List of AssociationParameter",
            Value="-", Description="-", Required="false", UpdateType="Mutable",
        ),
        PropertyField(
            Property="&nbsp;&nbsp;&nbsp;&nbsp;Key", Type="String",
            Value="KEY2", Description="-", Required="true", UpdateType="Mutable",
        ),
        PropertyField(
            Property="&nbsp;&nbsp;&nbsp;&nbsp;Value", Type="List of String",
            Value='[<br/>&nbsp;&nbsp;"VALUE21",<br/>&nbsp;&nbsp;"VALUE22"<br/>]', Description="-", Required="true", UpdateType="Mutable",
        ),
    ]

    generator = CfnMarkdownDocumentGenerator(context=context,)

    properties = generator._flatten_properties_node( # type: ignore
        cfn_template_tree.group_nodes[cfn_template_tree.group_name_for_independent_resources].resource_nodes["Instance"].properties_node,
    )
    assert len(properties) == len(expected)
    for p, e in zip(properties, expected):
        assert p.Property == e.Property
        assert p.Type == e.Type
        assert p.Value == e.Value
        assert p.Description == e.Description
        assert p.Required == e.Required
        assert p.UpdateType == e.UpdateType, p


@pytest.mark.parametrize("jsonpath,expected", [
    ("$.BlockDeviceMappings", "BlockDeviceMappings"),
    ("$.BlockDeviceMappings[0]", "BlockDeviceMappings[0]"),
    ("$.BlockDeviceMappings[0].Ebs", "&nbsp;&nbsp;Ebs"),
    ("$.BlockDeviceMappings[0].Ebs.Encrypted", "&nbsp;&nbsp;&nbsp;&nbsp;Encrypted"),
    ("$.SsmAssociations[0].AssociationParameters[0]", "&nbsp;&nbsp;AssociationParameters[0]"),
    ("$.SsmAssociations[0].AssociationParameters[0].Key", "&nbsp;&nbsp;&nbsp;&nbsp;Key"),
])
def test_CfnMarkdownDocumentGenerator_simplify_jsonpath(
    jsonpath:str, expected:str,
    context:AppContext,
):
    generator = CfnMarkdownDocumentGenerator(context=context,)
    path = generator._simplify_jsonpath(jsonpath) # type: ignore
    assert path == expected


@pytest.mark.parametrize("outputs_node,e_outputs", [
    (
        CfnTemplateOutputsNode(
            context=AppContext(log_level=logging.DEBUG),
            definitions={
                "out1": CfnTemplateOutputDefinition(
                    Description="out1-description",
                    Value={"Ref": "VALUE"},
                    Export=CfnTemplateOutputExportDefinition(
                        Name={"Ref": "NAME"}
                    ),
                    Condition="cond1"
                ),
                "out2": CfnTemplateOutputDefinition(
                    Value="VALUE"
                )
            }
        ),
        "\n".join([
            "## Outputs",
            "",
            "### out1",
            "",
            "out1-description",
            "",
            "|Value|ExportName|Condition|",
            "|-|-|-|",
            '|{<br/>&nbsp;&nbsp;"Ref":&nbsp;"VALUE"<br/>}|{<br/>&nbsp;&nbsp;"Ref":&nbsp;"NAME"<br/>}|cond1|',
            "",
            "### out2",
            "",
            "",
            "",
            "|Value|ExportName|Condition|",
            "|-|-|-|",
            "|VALUE|-|-|",
        ])
    ),
    (
        CfnTemplateOutputsNode(
            context=AppContext(log_level=logging.DEBUG),
            definitions={}
        ),
        "\n".join([
            "## Outputs",
        ])
    )
])
def test_CfnMarkdownDocumentGenerator_outputs(
    outputs_node:CfnTemplateOutputsNode, e_outputs:str,
    cfn_template_tree:CfnTemplateTree,
    context:AppContext,
):
    cfn_template_tree.outputs_node = outputs_node
    generator = CfnMarkdownDocumentGenerator(context=context)
    outputs = generator.outputs(cfn_template_tree)
    assert outputs == e_outputs

@pytest.mark.parametrize("tree,e_template", [
    (
        CfnTemplateTree(
            context=AppContext(log_level=logging.DEBUG),
            template_source=CfnTemplateSource("template.yaml", context=AppContext(log_level=logging.DEBUG)),
            definition=CfnTemplateDefinition(
                Resources={"resource1": CfnTemplateResourceDefinition(
                    Type="AWS::EC2::Instance",
                    Properties={},
                )}
            ),
            spec_repository=CfnSpecificationRepository(
                context=AppContext(log_level=logging.DEBUG),
                loader=RemoteFileLoader(context=AppContext(log_level=logging.DEBUG)),
                source_url=AppConfig.DEFAULT_SPECIFICATION_URL,
                cache=LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=AppContext(log_level=logging.DEBUG)),
                recursive_resource_types=AppConfig.RECURSIVE_RESOURCE_TYPES,
            )
        ),
        "\n".join([
            "- [template.yaml](#templateyaml)",
            "  - [Description](#description)",
            "  - [Parameters](#parameters)",
            "  - [Mappings](#mappings)",
            "  - [Conditions](#conditions)",
            "  - [Rules](#rules)",
            "  - [Resources](#resources)",
            "    - [resource1 (AWS::EC2::Instance)](#resource1-awsec2instance)",
            "  - [Outputs](#outputs)",
            "",
            "---",
            "",
            "# template.yaml",
            "",
            "| | |",
            "|-|-|",
            "|AWSTemplateFormatVersion|-|",
            "|Description|-|",
            "|Transform|-|",
            "",
            "---",
            "",
            "## Description",
            "",
            "",
            "---",
            "",
            "## Parameters",
            "",
            "---",
            "",
            "## Mappings",
            "",
            "---",
            "",
            "## Conditions",
            "",
            "---",
            "",
            "## Rules",
            "",
            "---",
            "",
            "## Resources",
            "",
            "### [resource1 (AWS::EC2::Instance)](http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ec2-instance.html)",
            "",
            "",
            "",
            "|DependsOn|Condition|CreationPolicy|UpdatePolicy|UpdateReplacePolicy|DeletionPolicy|",
            "|-|-|-|-|-|-|",
            "|-|-|-|-|Delete|Delete|",
            "",
            "|Property|Value|Description|Type|Required|UpdateType|",
            "|-|-|-|-|-|-|",
            "",
            "---",
            "",
            "## Outputs",
        ])
    )
])
def test_CfnMarkdownDocumentGenerator_generate(
    tree:CfnTemplateTree, e_template:str,
    context:AppContext,
):
    generator = CfnMarkdownDocumentGenerator(context=context)
    template = generator.generate(tree)
    header = generator.header()

    assert template == header + "\n" + e_template




from __future__ import annotations
from collections import defaultdict
from dataclasses import dataclass
import itertools
import json
import os
from jsonpath_ng import parse # type: ignore

from typing import Any, List, Literal, Mapping, Optional, Union, cast
from pydantic import BaseModel, Field, PositiveInt
from cfn_docgen.config import AppContext
from cfn_docgen.domain.model.cfn_specification import CfnSpecificationForResource, CfnSpecificationProperty, CfnSpecificationPropertyTypeName, CfnSpecificationResourceTypeName, CfnSpecificationPropertyType

from cfn_docgen.domain.ports.cfn_specification_repository import ICfnSpecificationRepository

class CfnTemplateParameterDefinition(BaseModel):
    """based on https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/parameters-section-structure.html"""
    AllowedPattern: Optional[str] = None
    AllowedValues: Optional[List[str] | List[int]] = None
    ConstraintDescription: Optional[str] = None
    Default: Optional[str | int | List[str] | List[int]] = None
    Description: Optional[str] = None
    MaxLength: Optional[PositiveInt] = None
    MinLength: Optional[PositiveInt] = None
    MaxValue: Optional[int] = None
    MinValue: Optional[int] = None
    NoEcho: bool = False
    Type: Literal[
        "String", "Number", "List<Number>", "CommaDelimitedList",
        "AWS::EC2::AvailabilityZone::Name",
        "AWS::EC2::Image::Id",
        "AWS::EC2::Instance::Id",
        "AWS::EC2::KeyPair::KeyName",
        "AWS::EC2::SecurityGroup::GroupName",
        "AWS::EC2::SecurityGroup::Id",
        "AWS::EC2::Subnet::Id",
        "AWS::EC2::Volume::Id",
        "AWS::EC2::VPC::Id",
        "AWS::Route53::HostedZone::Id",
        "List<AWS::EC2::AvailabilityZone::Name>",
        "List<AWS::EC2::Image::Id>",
        "List<AWS::EC2::Instance::Id>",
        "List<AWS::EC2::SecurityGroup::GroupName>",
        "List<AWS::EC2::SecurityGroup::Id>",
        "List<AWS::EC2::Subnet::Id>",
        "List<AWS::EC2::Volume::Id>",
        "List<AWS::EC2::VPC::Id>",
        "List<AWS::Route53::HostedZone::Id>",
        "AWS::SSM::Parameter::Name",
        "AWS::SSM::Parameter::Value<String>",
        "AWS::SSM::Parameter::Value<List<String>>",
        "AWS::SSM::Parameter::Value<CommaDelimitedList>",
        "AWS::SSM::Parameter::Value<AWS::EC2::AvailabilityZone::Name>",
        "AWS::SSM::Parameter::Value<AWS::EC2::Image::Id>",
        "AWS::SSM::Parameter::Value<AWS::EC2::Instance::Id>",
        "AWS::SSM::Parameter::Value<AWS::EC2::KeyPair::KeyName>",
        "AWS::SSM::Parameter::Value<AWS::EC2::SecurityGroup::GroupName>",
        "AWS::SSM::Parameter::Value<AWS::EC2::SecurityGroup::Id>",
        "AWS::SSM::Parameter::Value<AWS::EC2::Subnet::Id>",
        "AWS::SSM::Parameter::Value<AWS::EC2::Volume::Id>",
        "AWS::SSM::Parameter::Value<AWS::EC2::VPC::Id>",
        "AWS::SSM::Parameter::Value<AWS::Route53::HostedZone::Id>",
        "AWS::SSM::Parameter::Value<List<AWS::EC2::AvailabilityZone::Name>>"
        "AWS::SSM::Parameter::Value<List<AWS::EC2::Image::Id>>"
        "AWS::SSM::Parameter::Value<List<AWS::EC2::Instance::Id>>"
        "AWS::SSM::Parameter::Value<List<AWS::EC2::KeyPair::KeyName>>"
        "AWS::SSM::Parameter::Value<List<AWS::EC2::SecurityGroup::GroupName>>"
        "AWS::SSM::Parameter::Value<List<AWS::EC2::SecurityGroup::Id>>"
        "AWS::SSM::Parameter::Value<List<AWS::EC2::Subnet::Id>>"
        "AWS::SSM::Parameter::Value<List<AWS::EC2::Volume::Id>>"
        "AWS::SSM::Parameter::Value<List<AWS::EC2::VPC::Id>>"
        "AWS::SSM::Parameter::Value<List<AWS::Route53::HostedZone::Id>>"
    ]

class CfnTemplateMetadataCfnDocgenDefinition(BaseModel):
    Description: Optional[str] = None
    Mappings: Mapping[str, str] = {}
    Conditions: Mapping[str, str] = {}
    Rules: Mapping[str, str] = {}

class CfnTemplateMetadataParameterGroupLabel(BaseModel):
    default: str
class CfnTemplateMetadataParameterGroup(BaseModel):
    Label: CfnTemplateMetadataParameterGroupLabel
    Parameters: List[str]

class CfnTemplateMetadataInterface(BaseModel):
    ParameterGroups: List[CfnTemplateMetadataParameterGroup] = []
    ParameterLabels: Mapping[str, CfnTemplateMetadataParameterGroupLabel] = {}

class CfnTemplateMetadataDefinition(BaseModel):
    aws_cloudformation_interface:Optional[CfnTemplateMetadataInterface] = Field(default=None, alias="AWS::CloudFormation::Interface")
    CfnDocgen: Optional[CfnTemplateMetadataCfnDocgenDefinition] = None


class CfnTemplateRuleAssertDefinition(BaseModel):
    """based on https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/rules-section-structure.html"""
    Assert: Mapping[Any, Any]
    AssertDescription: str

class CfnTemplateRuleDefinition(BaseModel):
    """based on https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/rules-section-structure.html"""
    RuleCondition: Optional[Mapping[Any, Any]] = None
    Assertions: List[CfnTemplateRuleAssertDefinition]

# class CfnTemplateMappingDefinition(BaseModel):
#     """based on https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/mappings-section-structure.html"""

class CfnTemplateResourceMetadataCfnDocgenDefinition(BaseModel):
    Description: Optional[str] = None
    Properties: Optional[Mapping[str, Any]] = None


    def get_resource_description(self, ) -> Optional[str]:
        return self.Description

    def get_property_description_by_json_path(self, json_path:str) -> Optional[str]:
        if self.Properties is None:
            return None
        jsonpath_expr = parse(json_path) # type: ignore
        descriptions:List[Any] = [f.value for f in jsonpath_expr.find(self.Properties)] # type: ignore
        if len(descriptions) != 1:
            return None
        description = descriptions[0]
        if not isinstance(description, str):
            return None
        return description

class CfnTemplateResourceMetadataDefinition(BaseModel):
    CfnDocgen: Optional[CfnTemplateResourceMetadataCfnDocgenDefinition] = None
    aws_cdk_path: Optional[str] = Field(alias="aws:cdk:path", default=None)

CfnTemplateResourcePropertyDefinition = Union[
    str,int,float, bool, List[Any], Mapping[Any, Any]
]

class CfnTemplateResourceDefinition(BaseModel):
    """based on https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/resources-section-structure.html and https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-product-attribute-reference.html"""
    Type: str
    Metadata: Optional[CfnTemplateResourceMetadataDefinition] = None
    DependsOn: Optional[List[str]] = None
    Condition: Optional[str] = None
    Properties: Mapping[str, CfnTemplateResourcePropertyDefinition]
    CreationPolicy: Optional[Mapping[str, Any]] = None
    DeletionPolicy: Literal["Delete", "Retain", "RetainExceptOnCreate", "Snapshot"] = "Delete"
    UpdatePolicy: Optional[Mapping[str, Any]] = None
    UpdateReplacePolicy: Literal["Delete", "Retain", "Snapshot"] = "Delete"

    
class CfnTemplateOutputExportDefinition(BaseModel):
    Name: str | Mapping[str, Any]

class CfnTemplateOutputDefinition(BaseModel):
    """based on https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/outputs-section-structure.html"""
    Description: Optional[str] = None
    Value: Any
    Export: Optional[CfnTemplateOutputExportDefinition] = None
    Condition: Optional[str] = None

class CfnTemplateDefinition(BaseModel):
    """based on https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/template-anatomy.html"""
    Description: Optional[str] = None
    AWSTemplateFormatVersion: Optional[str] = None
    Metadata: Optional[CfnTemplateMetadataDefinition] = None
    Parameters: Mapping[str, CfnTemplateParameterDefinition] = {}
    Rules: Mapping[str, CfnTemplateRuleDefinition] = {}
    Mappings: Mapping[str, Mapping[str, Mapping[str, Any]]] = {}
    Conditions: Mapping[str, Any] = {}
    Transform: str | List[str] = []
    Resources: Mapping[str, CfnTemplateResourceDefinition]
    Outputs: Mapping[str, CfnTemplateOutputDefinition] = {}

    @classmethod
    def from_string(cls, template:str, context:AppContext) -> CfnTemplateDefinition:
        try:
            return CfnTemplateDefinition(**json.loads(template))
        except Exception as ex:
            context.log_error(f"failed to instanciate CfnTemplateDefinition from string. {str(ex)}")
            raise ex

    def __get_cfn_docgen(self) -> Optional[CfnTemplateMetadataCfnDocgenDefinition]:
        metadata = self.Metadata
        if metadata is None:
            return None
        cfn_docgen = metadata.CfnDocgen
        return cfn_docgen

    def get_parameter_groups(self) -> List[CfnTemplateMetadataParameterGroup]:
        metadata = self.Metadata
        if metadata is None:
            return []
        interface = metadata.aws_cloudformation_interface
        if interface is None:
            return []
        return interface.ParameterGroups
    
    def get_cfn_docgen_description(self) -> Optional[str]:
        cfn_docgen = self.__get_cfn_docgen()
        if cfn_docgen is None:
            return None
        return cfn_docgen.Description
    def get_mapping_descriptions(self) -> Mapping[str, str]:
        cfn_docgen = self.__get_cfn_docgen()
        if cfn_docgen is None:
            return {}
        return cfn_docgen.Mappings
    def get_rule_descriptions(self) -> Mapping[str, str]:
        cfn_docgen = self.__get_cfn_docgen()
        if cfn_docgen is None:
            return {}
        return cfn_docgen.Rules
    def get_condition_descriptions(self) -> Mapping[str, str]:
        cfn_docgen = self.__get_cfn_docgen()
        if cfn_docgen is None:
            return {}
        return cfn_docgen.Conditions
    
    def get_resource_groups(self) -> Mapping[str, List[str]]:
        
        resource_groups:Mapping[str, List[str]] = defaultdict(list)
        for resource_id, resource_body in self.Resources.items():
            if resource_body.Metadata is None:
                continue
            if resource_body.Metadata.aws_cdk_path is None:
                continue
            # aws:cdk:path wll be like CdkWorkshopStack/BlogVpc/PrivateSubnet1/Subnet
            paths = resource_body.Metadata.aws_cdk_path.split("/")
            if len(paths) < 2:
                continue
            resource_groups[paths[1]].append(resource_id)

        return resource_groups


class CfnTemplateParametersNode:

    group_name_for_independent_parameters = "__CFN_DOCGEN_INDEPENDENT_PARAMETERS__"

    def __init__(
        self,
        definitions:Mapping[str, CfnTemplateParameterDefinition],
        parameter_groups:List[CfnTemplateMetadataParameterGroup],
        context:AppContext,
    ) -> None:
        
        self.group_nodes:Mapping[str, CfnTemplateParameterGroupNode] ={}
        for group in parameter_groups:
            try:
                self.group_nodes[group.Label.default] = CfnTemplateParameterGroupNode(
                    definitions={
                        param_name: definition 
                        for param_name, definition in definitions.items()
                        if param_name in group.Parameters
                    },
                    context=context,
                )
            except Exception:
                context.log_warning(f"failed to build CfnTemplateParameterGroupNode for parameter group [{group.Label.default}]")
                continue

        try:
            parameter_names_in_groups = list(itertools.chain.from_iterable(
                [group.Parameters for group in parameter_groups]
            ))
            self.group_nodes[self.group_name_for_independent_parameters] = CfnTemplateParameterGroupNode(
                definitions={
                    param_name: definition for param_name, definition in definitions.items()
                    if param_name not in parameter_names_in_groups
                },
                context=context,
            )
        except Exception:
            context.log_warning(f"failed to build CfnTemplateParameterGroupNode for parameter group [{self.group_name_for_independent_parameters}]")


class CfnTemplateParameterGroupNode:
    def __init__(
        self,
        definitions:Mapping[str, CfnTemplateParameterDefinition],
        context:AppContext,
    ) -> None:
        self.leaves:Mapping[str, CfnTemplateParameterLeaf] = {}
        for name, definition in definitions.items():
            try:
                self.leaves[name] = CfnTemplateParameterLeaf(definition=definition)
            except Exception:
                context.log_warning(f"failed to build CfnTemplateParameterLeaf for parameter [{name}]")
                continue

class CfnTemplateParameterLeaf:
    def __init__(self, definition:CfnTemplateParameterDefinition) -> None:
        self.definition = definition


class CfnTemplateMappingLeaf:
    def __init__(
        self, 
        definition:Mapping[str, Mapping[str, Any]],
        description:Optional[str],
    ) -> None:
        self.definition = definition
        self.description = description

class CfnTemplateMappingsNode:
    def __init__(
        self,
        definitions:Mapping[str, Mapping[str, Mapping[str, Any]]],
        descriptions:Mapping[str, str],
        context:AppContext,
    ) -> None:
        self.mapping_leaves:Mapping[str, CfnTemplateMappingLeaf] = {}
        for name, definition in definitions.items():
            try:
                self.mapping_leaves[name] = CfnTemplateMappingLeaf(
                    definition=definition,
                    description=descriptions.get(name)
                )
            except Exception:
                context.log_warning(f"failed to build CfnTemplateMappingLeaf for mapping [{name}]")
                continue

class CfnTemplateConditionLeaf:
    def __init__(
        self,
        definition: Mapping[str, Any],
        description: Optional[str],
    ) -> None:
        self.definition = definition
        self.descirption = description

class CfnTemplateConditionsNode:
    def __init__(
        self,
        definitions: Mapping[str, Mapping[str, Any]],
        descriptions: Mapping[str, str],
        context:AppContext,
    ) -> None:
        self.condition_leaves:Mapping[str, CfnTemplateConditionLeaf] = {}
        for name, definition in definitions.items():
            try:
                self.condition_leaves[name] = CfnTemplateConditionLeaf(
                    definition=definition,
                    description=descriptions.get(name)
                )
            except Exception:
                context.log_warning(f"failed to build CfnTemplateConditionLeaf for condition [{name}]")
                continue

class CfnTemplateRuleLeaf:
    def __init__(
        self,
        definition: CfnTemplateRuleDefinition,
        description: Optional[str]
    ) -> None:
        self.definition = definition
        self.description = description


class CfnTemplateRulesNode:
    def __init__(
        self,
        definitions:Mapping[str, CfnTemplateRuleDefinition],
        descriptions:Mapping[str, str],
        context:AppContext,
    ) -> None:
        self.rule_leaves:Mapping[str, CfnTemplateRuleLeaf] = {}
        for name, definition in definitions.items():
            try:
                self.rule_leaves[name] = CfnTemplateRuleLeaf(
                    definition=definition,
                    description=descriptions.get(name),
                )
            except Exception:
                context.log_warning(f"failed to build CfnTemplateRuleLeaf for rule [{name}]")
                continue

class CfnTemplateResourcePropertyLeaf:
    def __init__(
        self,
        definition:Optional[CfnTemplateResourcePropertyDefinition],
        property_spec:CfnSpecificationProperty,
        description:Optional[str],
        json_path:str,
    ) -> None:
        self.definition = definition
        self.spec = property_spec
        self.description = description
        self.json_path = json_path



class CfnTemplateResourcePropertyNode:
    def __init__(
        self,
        definitions:Optional[CfnTemplateResourcePropertyDefinition],
        resource_property_spec:Optional[CfnSpecificationProperty],
        description:Optional[str],
        properties_spec:Mapping[str, CfnSpecificationProperty],
        json_path:str,
        cfn_docgen_metadata:Optional[CfnTemplateResourceMetadataCfnDocgenDefinition],
        property_specs:Mapping[str, CfnSpecificationPropertyType],
        resource_info:ResourceInfo,
        context:AppContext,
    ) -> None:
        self.property_nodes:Mapping[
            str, # e.g. CpuOptions in AWS::EC2::Instance
            CfnTemplateResourcePropertyNode # its node
        ] = {}
        self.property_nodes_map:Mapping[
            str, # e.g. CustomResponseBodies in AWS::WAFv2::WebACL
            Mapping[
                str, # its key
                CfnTemplateResourcePropertyNode # its node
            ]
        ] = {}
        self.property_nodes_list:Mapping[
            str, # e.g. BlockDeviceMappings in AWS::EC2::Instance
            List[CfnTemplateResourcePropertyNode] # its list of node
        ] = {}
        self.property_leaves:Mapping[
            str, # e.g. ImageId in AWS::EC2::Instance
            CfnTemplateResourcePropertyLeaf # its definition
        ] = {}

        self.description = description
        self.json_path = json_path
        self.spec = resource_property_spec
        self.has_leaves = definitions is not None


        for property_name, property_spec in properties_spec.items():
            try:
                if (
                    property_spec.PrimitiveType is not None 
                    or property_spec.PrimitiveItemType is not None
                ):
                # e.g. CoreCount in AWS::EC2::Instance.CpuOptions (string primitive type)
                # or Value in AWS::EC2::Instance.AssociationParameter (list of string primitive item type)
                # or Options in AWS::ECS::Service.LogConfiguration (map of string primitive item type)
                    definition:Any = None
                    if definitions is not None and isinstance(definitions, dict):
                        definition = cast(Mapping[str, CfnTemplateResourcePropertyDefinition], definitions).get(property_name, None)
                    prop_json_path = f"{self.json_path}.{property_name}"
                    description = cfn_docgen_metadata.get_property_description_by_json_path(
                        json_path=prop_json_path,
                    ) if cfn_docgen_metadata is not None else None
                    if resource_info.is_recursive and definition is None:
                        continue
                    self.property_leaves[property_name] = CfnTemplateResourcePropertyLeaf(
                        definition=definition,
                        property_spec=property_spec,
                        description=description,
                        json_path=prop_json_path,
                    )
                    continue

            except Exception:
                context.log_warning(f"failed to build CfnTemplateResourcePropertyLeaf for property [{property_name}]")
                continue

            try:

                if property_spec.ItemType is not None:
                    if (
                        property_spec.Type is not None
                        and property_spec.Type.lower() == "list"
                    ):
                    # e.g. AssociationParameters in AWS::EC2::Instance.SsmAssociation
                        _property_spec = property_specs.get(
                            CfnSpecificationPropertyTypeName(f"{resource_info.type}.{property_spec.ItemType}", context).fullname
                        )
                        if _property_spec is None or _property_spec.Properties is None:
                            continue
                        definition_list:List[Any] = []
                        if definitions is not None and isinstance(definitions, dict):
                            definition_list = cast(Mapping[str, List[Any]], definitions).get(property_name, [None])
                        if not isinstance(definition_list, list): # type: ignore
                            raise ValueError
                        property_nodes_list:List[CfnTemplateResourcePropertyNode] = []
                        for i, definition in enumerate(definition_list):
                            # prop_name is Key, Value, ...
                            prop_json_path = f"{self.json_path}.{property_name}[{i}]"
                            description = cfn_docgen_metadata.get_property_description_by_json_path(
                                json_path=prop_json_path,
                            ) if cfn_docgen_metadata is not None else None
                            if resource_info.is_recursive and definition is None:
                                continue
                            property_nodes_list.append(CfnTemplateResourcePropertyNode(
                                definitions=definition,
                                resource_property_spec=property_spec,
                                description=description,
                                json_path=prop_json_path,
                                properties_spec=_property_spec.Properties,
                                property_specs=property_specs,
                                cfn_docgen_metadata=cfn_docgen_metadata,
                                resource_info=resource_info,
                                context=context,
                            ))
                        self.property_nodes_list[property_name] = property_nodes_list.copy()
                        continue

                    if (
                        property_spec.Type is not None
                        and property_spec.Type.lower() == "map"
                    ):
                    # e.g. RequestBodyAssociatedResourceTypeConfig in AWS::WAFv2::WebACL.AssociationConfig
                        
                        _property_spec = property_specs.get(
                            CfnSpecificationPropertyTypeName(f"{resource_info.type}.{property_spec.ItemType}", context).fullname
                        )
                        if _property_spec is None or _property_spec.Properties is None:
                            continue
                        definition_map:Mapping[str, Any] = {}
                        if definitions is not None and isinstance(definitions, dict):
                            definition_map = cast(Mapping[str, Mapping[str, Any]], definitions).get(property_name, {"key": None})
                        if not isinstance(definition_map, dict): # type: ignore
                            raise ValueError
                        property_nodes_map:Mapping[str, CfnTemplateResourcePropertyNode] = {}
                        for key, definition in definition_map.items():
                            # prop_name is DefaultSizeInspectionLimit, ...
                            prop_json_path = f"{self.json_path}.{property_name}.{key}"
                            description = cfn_docgen_metadata.get_property_description_by_json_path(
                                json_path=prop_json_path
                            ) if cfn_docgen_metadata is not None else None
                            if resource_info.is_recursive and definition is None:
                                continue
                            property_nodes_map[key] = CfnTemplateResourcePropertyNode(
                                definitions=definition,
                                resource_property_spec=property_spec,
                                properties_spec=_property_spec.Properties,
                                property_specs=property_specs,
                                description=description,
                                json_path=prop_json_path,
                                cfn_docgen_metadata=cfn_docgen_metadata,
                                resource_info=resource_info,
                                context=context,
                            )
                        self.property_nodes_map[property_name] = property_nodes_map.copy()
                        continue

                # the rest is e.g. Ebs in AWS::EC2::Instance.BlockDeviceMapping
                _property_spec = property_specs.get(
                    CfnSpecificationPropertyTypeName(f"{resource_info.type}.{property_spec.Type}", context).fullname
                )
                if _property_spec is None and property_spec.Type == "Tag":
                    _property_spec = property_specs.get(
                        CfnSpecificationPropertyTypeName(property_spec.Type, context).fullname
                    )
                if _property_spec is None or _property_spec.Properties is None:
                    continue
                definition:Any = None
                if definitions is not None and isinstance(definitions, dict):
                    definition = cast(Mapping[str, Any], definitions).get(property_name, None)
                prop_json_path = f"{self.json_path}.{property_name}"
                description = cfn_docgen_metadata.get_property_description_by_json_path(
                    json_path=prop_json_path
                ) if cfn_docgen_metadata is not None else None
                if resource_info.is_recursive and definition is None:
                    continue
                self.property_nodes[property_name] = CfnTemplateResourcePropertyNode(
                    definitions=definition,
                    resource_property_spec=property_spec,
                    properties_spec=_property_spec.Properties,
                    property_specs=property_specs,
                    cfn_docgen_metadata=cfn_docgen_metadata,
                    resource_info=resource_info,
                    json_path=prop_json_path,
                    description=description,
                    context=context,
                )
                continue

            except Exception:
                context.log_warning(f"failed to build CfnTemplateResourcePropertyNode for property [{property_name}]")
                continue


class CfnTemplateResourcePropertiesNode:
    def __init__(
        self,
        definitions:Mapping[str, CfnTemplateResourcePropertyDefinition],
        resource_property_specs:Mapping[str, CfnSpecificationProperty],
        property_specs:Mapping[str, CfnSpecificationPropertyType],
        cfn_docgen_metadata:Optional[CfnTemplateResourceMetadataCfnDocgenDefinition],
        resource_info:ResourceInfo,
        context:AppContext,

    ) -> None:
        self.property_nodes:Mapping[
            str, # e.g. CpuOptions in AWS::EC2::Instance
            CfnTemplateResourcePropertyNode # its node
        ] = {}
        self.property_nodes_map:Mapping[
            str, # e.g. CustomResponseBodies in AWS::WAFv2::WebACL
            Mapping[
                str, # its key
                CfnTemplateResourcePropertyNode # its node
            ]
        ] = {}
        self.property_nodes_list:Mapping[
            str, # e.g. BlockDeviceMappings in AWS::EC2::Instance
            List[CfnTemplateResourcePropertyNode] # its list of node
        ] = {}

        self.property_leaves:Mapping[
            str, # e.g. ImageId in AWS::EC2::Instance
            CfnTemplateResourcePropertyLeaf # its definition
        ] = {}

        self.json_path = "$"

        for property_name, resource_property_spec in resource_property_specs.items():
        # property_name is like ImageId, BlockDeviceMappings, CpuOptions, ... in AWS::EC2::Instance

            try:
                if (
                    resource_property_spec.PrimitiveType is not None 
                    or resource_property_spec.PrimitiveItemType is not None
                ):
                # e.g. ImageId in AWS::EC2::Instance (string primitive type)
                # or SecurityGroupIds in AWS::EC2::Instance (list of string primitive item type)
                # or Tags in AWS::MSK::Cluster (map of string primitive item type)
                    definition = definitions.get(property_name, None)
                    prop_json_path = f"{self.json_path}.{property_name}"
                    description = cfn_docgen_metadata.get_property_description_by_json_path(
                        json_path=prop_json_path,
                    ) if cfn_docgen_metadata is not None else None
                    if resource_info.is_recursive and definition is None:
                        continue
                    self.property_leaves[property_name] = CfnTemplateResourcePropertyLeaf(
                        definition=definition,
                        property_spec=resource_property_spec,
                        description=description,
                        json_path=prop_json_path,
                    )
                    continue
            except Exception:
                context.log_warning(f"failed to build CfnTemplateResourcePropertyLeaf for property [{property_name}]")
                continue

            try:
                if resource_property_spec.ItemType is not None:
                    if (
                        resource_property_spec.Type is not None
                        and resource_property_spec.Type.lower() == "list"
                    ):
                    # e.g. BlockDeviceMappings in AWS::EC2::Instance
                        property_spec = property_specs.get(
                            CfnSpecificationPropertyTypeName(f"{resource_info.type}.{resource_property_spec.ItemType}", context).fullname
                        )
                        if property_spec is None and resource_property_spec.ItemType == "Tag":
                            property_spec = property_specs.get(
                                CfnSpecificationPropertyTypeName(resource_property_spec.ItemType, context).fullname
                            )
                        if property_spec is None or property_spec.Properties is None:
                            continue
                        definition_list = definitions.get(property_name, [None])
                        if not isinstance(definition_list, list):
                            raise ValueError
                        property_nodes_list:List[CfnTemplateResourcePropertyNode] = []
                        for i, definition in enumerate(definition_list):
                            # prop_name is DeviceName, Ebs, ...
                            prop_json_path = f"{self.json_path}.{property_name}[{i}]"
                            description = cfn_docgen_metadata.get_property_description_by_json_path(
                                json_path=prop_json_path,
                            ) if cfn_docgen_metadata is not None else None
                            if resource_info.is_recursive and definition is None:
                                continue
                            property_nodes_list.append(CfnTemplateResourcePropertyNode(
                                definitions=definition,
                                resource_property_spec=resource_property_spec,
                                description=description,
                                json_path=prop_json_path,
                                properties_spec=property_spec.Properties,
                                property_specs=property_specs,
                                resource_info=resource_info,
                                cfn_docgen_metadata=cfn_docgen_metadata,
                                context=context,
                            ))
                        self.property_nodes_list[property_name] = property_nodes_list.copy()
                        continue

                    if (
                        resource_property_spec.Type is not None
                        and resource_property_spec.Type.lower() == "map"
                    ):
                    # e.g. CustomResponseBody in AWS::WAFv2::WebACL
                        
                        property_spec = property_specs.get(
                            CfnSpecificationPropertyTypeName(f"{resource_info.type}.{resource_property_spec.ItemType}", context).fullname
                        )
                        if property_spec is None or property_spec.Properties is None:
                            continue
                        definition_map = definitions.get(property_name, {"key": None})
                        if not isinstance(definition_map, dict):
                            raise ValueError
                        property_nodes_map:Mapping[str, CfnTemplateResourcePropertyNode] = {}
                        for key, definition in definition_map.items():
                            # for prop_name, prop_spec in property_spec.Properties.items():
                            # prop_name is ContentType, Content, ...
                            prop_json_path = f"{self.json_path}.{property_name}.{key}"
                            description = cfn_docgen_metadata.get_property_description_by_json_path(
                                json_path=prop_json_path
                            ) if cfn_docgen_metadata is not None else None
                            if resource_info.is_recursive and definition is None:
                                continue
                            property_nodes_map[key] = CfnTemplateResourcePropertyNode(
                                definitions=definition,
                                resource_property_spec=resource_property_spec,
                                description=description,
                                json_path=prop_json_path,
                                properties_spec=property_spec.Properties,
                                property_specs=property_specs,
                                resource_info=resource_info,
                                cfn_docgen_metadata=cfn_docgen_metadata,
                                context=context,
                            )
                        self.property_nodes_map[property_name] = property_nodes_map.copy()
                        continue

                # the rest is e.g. CpuOptions in AWS::EC2::Instance
                property_spec = property_specs.get(
                    CfnSpecificationPropertyTypeName(f"{resource_info.type}.{resource_property_spec.Type}", context).fullname
                )
                if property_spec is None or property_spec.Properties is None:
                    continue
                definition = definitions.get(property_name, None)
                prop_json_path = f"{self.json_path}.{property_name}"
                description = cfn_docgen_metadata.get_property_description_by_json_path(
                    json_path=prop_json_path
                ) if cfn_docgen_metadata is not None else None
                if resource_info.is_recursive and definition is None:
                    continue
                self.property_nodes[property_name] = CfnTemplateResourcePropertyNode(
                    definitions=definition,
                    resource_property_spec=resource_property_spec,
                    properties_spec=property_spec.Properties,
                    property_specs=property_specs,
                    resource_info=resource_info,
                    json_path=prop_json_path,
                    description=description,
                    cfn_docgen_metadata=cfn_docgen_metadata,
                    context=context,
                )
                continue

            except Exception:
                context.log_warning(f"failed to build CfnTemplateResourcePropertyNode for property [{property_name}]")
                continue


class CfnTemplateResourceNode:
    def __init__(
        self,
        definition:CfnTemplateResourceDefinition,
        specs:CfnSpecificationForResource,
        resource_info:ResourceInfo,
        context:AppContext,
    ) -> None:
        metadata = definition.Metadata
        cfn_docgen = metadata.CfnDocgen if metadata is not None else None

        self.description:Optional[str] = cfn_docgen.get_resource_description() if cfn_docgen is not None else None
        self.condition = definition.Condition
        self.creation_policy = definition.CreationPolicy
        self.update_policy =  definition.UpdatePolicy
        self.deletion_policy = definition.DeletionPolicy
        self.update_replace_policy = definition.UpdateReplacePolicy
        self.metadata = definition.Metadata
        self.depends_on:List[str] = definition.DependsOn if definition.DependsOn is not None else []
        self.type = resource_info.type

        self.spec = specs.ResourceSpec

        try:
            self.properties_node = CfnTemplateResourcePropertiesNode(
                definitions=definition.Properties,
                resource_property_specs=specs.ResourceSpec.Properties,
                property_specs=specs.PropertySpecs,
                cfn_docgen_metadata=cfn_docgen,
                resource_info=resource_info,
                context=context,
            )
        except Exception:
            context.log_warning(f"failed to build CfnTemplateResourcePropertiesNode")


class CfnTemplateResourceGroupNode:
    def __init__(
        self,
        definitions:Mapping[str, CfnTemplateResourceDefinition],
        spec_repository:ICfnSpecificationRepository,
        context:AppContext,

    ) -> None:
        self.resource_nodes:Mapping[str, CfnTemplateResourceNode] = {}
        for name, definition in definitions.items():
            try:
                specs = spec_repository.get_specs_for_resource(CfnSpecificationResourceTypeName(definition.Type, context))
                resource_info=ResourceInfo(
                    type=definition.Type, 
                    is_recursive=spec_repository.is_recursive(CfnSpecificationResourceTypeName(definition.Type, context))
                )
                self.resource_nodes[name] = CfnTemplateResourceNode(
                    definition=definition, 
                    specs=specs,
                    resource_info=resource_info,
                    context=context,
                )
            except Exception:
                context.log_warning(f"failed to build CfnTemplateResourceNode for resource [{name}]")

@dataclass
class ResourceInfo:
    type: str
    is_recursive: bool

class CfnTemplateResourcesNode:
    group_name_for_independent_resources = "__CFN_DOCGEN_INDEPENDENT_RESOURCES__"

    def __init__(
        self,
        definitions:Mapping[str, CfnTemplateResourceDefinition],
        resource_groups:Mapping[str, List[str]],
        spec_repository:ICfnSpecificationRepository,
        context:AppContext,

    ) -> None:
        self.group_nodes:Mapping[str, CfnTemplateResourceGroupNode] = {}
        
        for group_name, resource_ids in resource_groups.items():
            try:
                self.group_nodes[group_name] = CfnTemplateResourceGroupNode(
                    definitions={
                        id_: d for id_, d in definitions.items()
                        if id_ in resource_ids
                    },
                    spec_repository=spec_repository,
                    context=context,
                )
            except Exception:
                context.log_warning(f"failed to build CfnTemplateResourcesNode for resource group [{group_name}]")
                continue
        try:
            resource_ids_in_groups = list(itertools.chain.from_iterable(
                [resource_ids for resource_ids in resource_groups.values()]
            ))
            self.group_nodes[self.group_name_for_independent_resources] = CfnTemplateResourceGroupNode(
                definitions={
                    id_: d for id_, d in definitions.items()
                    if id_ not in resource_ids_in_groups
                },
                spec_repository=spec_repository,
                context=context,
            )
        except Exception:
            context.log_warning(f"failed to build CfnTemplateResourcesNode for resource group [{self.group_name_for_independent_resources}]")

        # self.resource_nodes:Mapping[str, CfnTemplateResourceNode] = {}
        # for name, definition in definitions.items():
        #     try:
        #         specs = spec_repository.get_specs_for_resource(CfnSpecificationResourceTypeName(definition.Type, context))
        #         resource_info=ResourceInfo(
        #             type=definition.Type, 
        #             is_recursive=spec_repository.is_recursive(CfnSpecificationResourceTypeName(definition.Type, context))
        #         )
        #         self.resource_nodes[name] = CfnTemplateResourceNode(
        #             definition=definition, 
        #             specs=specs,
        #             resource_info=resource_info,
        #             context=context,
        #         )
        #     except Exception:
        #         context.log_warning(f"failed to build CfnTemplateResourceNode for resource [{name}]")

class CfnTemplateOutputLeaf:
    def __init__(
        self,
        definition:CfnTemplateOutputDefinition    
    ) -> None:
        self.definition = definition

class CfnTemplateOutputsNode:
    def __init__(
        self,
        definitions:Mapping[str, CfnTemplateOutputDefinition],
        context:AppContext,
    ) -> None:
        self.output_leaves:Mapping[str, CfnTemplateOutputLeaf] = {}
        for name, definition in definitions.items():
            try:
                self.output_leaves[name] = CfnTemplateOutputLeaf(
                    definition=definition,
                )
            except Exception:
                context.log_warning(f"failed to build CfnTemplateOutputLeaf for output [{name}]")


SupportedSourceType = Literal["LocalFilePath", "S3BucketKey", "HttpUrl"]
class CfnTemplateSource:
    type: SupportedSourceType
    source: str
    def __init__(self, source:str, context:AppContext) -> None:
        self.source = source
        if source.startswith("s3://"):
            self.type = "S3BucketKey"
        elif source.startswith("http://") or source.startswith("https://"):
            self.type = "HttpUrl"
        else:
            self.type = "LocalFilePath"

        context.log_debug(f"the type of template source [{self.source}] is [{self.type}]")
    
    @property
    def basename(self) -> str:
        return os.path.basename(self.source)


class CfnTemplateTree:
    """treet CfnTemplate as tree"""
    def __init__(
        self, 
        template_source:CfnTemplateSource,
        definition:CfnTemplateDefinition, 
        spec_repository:ICfnSpecificationRepository,
        context:AppContext,
    ) -> None:
        self.template_source = template_source
        self.description = definition.Description
        self.aws_template_format_version = definition.AWSTemplateFormatVersion
        self.transform = definition.Transform
        self.cfn_docgen_description = definition.get_cfn_docgen_description()

        self.parameters_node = CfnTemplateParametersNode(
            definitions=definition.Parameters,
            parameter_groups=definition.get_parameter_groups(),
            context=context,
        )
        self.mappings_node = CfnTemplateMappingsNode(
            definitions=definition.Mappings,
            descriptions=definition.get_mapping_descriptions(),
            context=context,
        )
        self.rules_node = CfnTemplateRulesNode(
            definitions=definition.Rules,
            descriptions=definition.get_rule_descriptions(),
            context=context,
        )
        self.conditions_node = CfnTemplateConditionsNode(
            definitions=definition.Conditions,
            descriptions=definition.get_condition_descriptions(),
            context=context,
        )
        self.resources_node = CfnTemplateResourcesNode(
            definitions=definition.Resources,
            resource_groups=definition.get_resource_groups(),
            spec_repository=spec_repository,
            context=context,
        )
        self.outputs_node = CfnTemplateOutputsNode(
            definitions=definition.Outputs,
            context=context,
        )




from dataclasses import dataclass
import os
from typing import Callable, Literal, List
from cfn_docgen.config import AppContext

from cfn_docgen.domain.model.cfn_document_generator import CfnDocumentDestination
from cfn_docgen.domain.model.cfn_template import CfnTemplateSource
from cfn_docgen.domain.ports.internal.file_loader import IFileLoader

from cfn_docgen.domain.services.cfn_docgen_service import CfnDocgenServiceCommandInput, SupportedFormat, ext_by_format


Subcommand = Literal["docgen"]


@dataclass
class CliArguement:
    subcommand:Subcommand
    format: SupportedFormat
    source: str
    dest: str
    debug: bool = False

    def as_list(self) -> List[str]:
        arg_list = [
            self.subcommand,
            "--format", self.format,
            "--source", self.source,
            "--dest", self.dest,
        ]
        if self.debug:
            arg_list.append("--debug")
        return arg_list
    
class CfnDocgenCLIUnitsOfWork:
    args:CliArguement
    units_of_work:List[CfnDocgenServiceCommandInput]

    def __init__(
        self,
        args:CliArguement,
        file_loader_factory:Callable[[CfnTemplateSource, AppContext], IFileLoader],
        context:AppContext,
    ) -> None:
        self.units_of_work = self.build_units_of_work(
            args=args,
            file_loader_factory=file_loader_factory,
            context=context,
        )
        try:
            assert len(self.units_of_work) > 0, "no valid template sources and document dests are provided"
        except AssertionError as ex:
            context.log_error(ex.args[0])
            raise ex

    def build_units_of_work(
        self, 
        args:CliArguement, 
        file_loader_factory:Callable[[CfnTemplateSource, AppContext], IFileLoader],
        context:AppContext,
    ) -> List[CfnDocgenServiceCommandInput]:
        try:
            loader = file_loader_factory(CfnTemplateSource(args.source, context=context), context)
            sources = loader.list(source=args.source)
            assert len(sources) > 0, f"failed to list template sources at [{args.source}]"
            context.log_debug(f"listed source template files [{'.'.join(sources)}]")
        except Exception as ex:
            context.log_error(str(ex))
            return []
        

        # determine dest type
        dests:List[str] = []
        try:
            # dest is a single file
            _, dest_ext = os.path.splitext(args.dest)
            if dest_ext[1:] in list(ext_by_format.values()):
                assert len(sources) == 1
                return [
                    CfnDocgenServiceCommandInput(
                        template_source=CfnTemplateSource(s, context=context),
                        document_dest=CfnDocumentDestination(args.dest, context=context),
                        fmt=args.format,
                    ) for s in sources
                ]
        except Exception:
            context.log_error(f"invalid arguement pattern: source {args.source} is a directory, while dest {args.dest} is a single file")
            return []

        try:
            # dest is a directory
            for s in sources:
                base_prefix, _ = os.path.splitext(os.path.basename(s))
                subdirs = s.replace(args.source, "")
                if subdirs.startswith("/") or subdirs.startswith("\\"):
                    subdirs = subdirs[1:]
                dests.append(
                    os.path.join(
                        args.dest, os.path.dirname(subdirs), f"{base_prefix}.{ext_by_format[args.format]}"
                    )
                )
            assert len(sources) == len(dests)
            units_of_work = [
                CfnDocgenServiceCommandInput(
                    template_source=CfnTemplateSource(s, context=context),
                    document_dest=CfnDocumentDestination(t, context=context),
                    fmt=args.format,
                ) for s, t in zip(sources, dests)
            ]
            units_of_work = sorted(
                units_of_work,
                key=lambda u: (u.template_source.source, u.document_dest.dest)
            )
            return units_of_work
        except Exception:
            context.log_error(f"the number of template sources[{len(sources)}] and document dests[{len(dests)}] are missmatched")
            return []


    def provide(self) -> List[CfnDocgenServiceCommandInput]:
        return self.units_of_work



from __future__ import annotations
from dataclasses import dataclass
import logging
from typing import Callable, Mapping
from cfn_docgen.adapters.cfn_document_storage import document_storage_facotory
from cfn_docgen.adapters.cfn_specification_repository import CfnSpecificationRepository
from cfn_docgen.adapters.cfn_template_provider import template_provider_factory
from cfn_docgen.adapters.internal.cache import LocalFileCache
from cfn_docgen.adapters.internal.file_loader import RemoteFileLoader
from cfn_docgen.config import AppConfig, AppContext, AwsConnectionSettings, ConnectionSettings
from cfn_docgen.domain.model.cfn_document_generator import CfnDocumentDestination, ICfnDocumentGenerator, SupportedFormat, document_generator_factory
from cfn_docgen.domain.model.cfn_template import CfnTemplateSource, CfnTemplateTree
from cfn_docgen.domain.ports.cfn_document_storage import ICfnDocumentStorage
from cfn_docgen.domain.ports.cfn_specification_repository import ICfnSpecificationRepository
from cfn_docgen.domain.ports.cfn_template_provider import ICfnTemplateProvider

@dataclass
class CfnDocgenServiceCommandInput:
    template_source:CfnTemplateSource
    document_dest:CfnDocumentDestination
    fmt: SupportedFormat
@dataclass
class CfnDocgenServiceCommandOutput:
    document_dest: str

ext_by_format:Mapping[SupportedFormat, str] = {
    "markdown": "md",
}


class CfnDocgenService(object):

    def __init__(
        self,
        cfn_template_provider_facotry:Callable[[CfnTemplateSource, AppContext], ICfnTemplateProvider],
        cfn_document_generator_factory:Callable[[SupportedFormat, AppContext], ICfnDocumentGenerator],
        cfn_document_storage_factory:Callable[[CfnDocumentDestination, AppContext], ICfnDocumentStorage],
        cfn_specification_repository:ICfnSpecificationRepository,
        context:AppContext,
    ) -> None:
        self.context = context
        try:
            self.template_provider_factory = cfn_template_provider_facotry
            self.document_generator_factory = cfn_document_generator_factory
            self.document_storage_factory = cfn_document_storage_factory
            self.spec_repository = cfn_specification_repository
        except Exception as ex:
            self.context.log_error("failed to setup cfn-docgen service")
            raise ex

    def main(self, command_input:CfnDocgenServiceCommandInput) -> CfnDocgenServiceCommandOutput:
        try:
            template_definition = self.template_provider_factory(
                command_input.template_source,
                self.context,
            ).load_template(
                command_input.template_source,
            )
        except Exception as ex:
            self.context.log_error(
                f"failed to load template from [{command_input.template_source.source}]"
            )
            raise ex

        try:
            template_tree = CfnTemplateTree(
                template_source=command_input.template_source,
                definition=template_definition,
                spec_repository=self.spec_repository,
                context=self.context,
            )
        except Exception as ex:
            self.context.log_error("failed to build CfnTemplateTree")
            raise ex

        try:
            document = self.document_generator_factory(
                command_input.fmt,
                self.context,
            ).generate(
                cfn_template_tree=template_tree,
            )
        except Exception as ex:
            self.context.log_error(f"failed to generate [{command_input.fmt}] document")
            raise ex
        try:
            self.document_storage_factory(
                command_input.document_dest,
                self.context,
            ).save_document(
                body=document.encode(encoding="UTF-8"),
                document_dest=command_input.document_dest,
            )
        except Exception as ex:
            self.context.log_error(f"failed to save [{command_input.fmt}] document")
            raise ex
        
        return CfnDocgenServiceCommandOutput(
            document_dest=command_input.document_dest.dest,
        )


    @classmethod
    def with_default(cls) -> CfnDocgenService:
        context = AppContext(
            log_level=logging.CRITICAL,
            connection_settings=ConnectionSettings(aws=AwsConnectionSettings(profile_name=None)),
        )
        return CfnDocgenService(
            context=context,
            cfn_template_provider_facotry=template_provider_factory,
            cfn_document_generator_factory=document_generator_factory,
            cfn_document_storage_factory=document_storage_facotory,
            cfn_specification_repository=CfnSpecificationRepository(
                context=context,
                source_url=AppConfig.DEFAULT_SPECIFICATION_URL,
                loader=RemoteFileLoader(context=context),
                cache=LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=context),
                recursive_resource_types=AppConfig.RECURSIVE_RESOURCE_TYPES,
            )
        )


import logging
import pytest
from cfn_docgen.config import AppContext, AwsConnectionSettings, ConnectionSettings
from cfn_docgen.domain.model.cfn_specification import CfnSpecificationPropertyTypeName, CfnSpecificationResourceTypeName

@pytest.fixture
def context():
    return AppContext(
        log_level=logging.DEBUG,
        connection_settings=ConnectionSettings(aws=AwsConnectionSettings(profile_name=None)),
    )

@pytest.mark.parametrize("resource_type", [
    ("AWS::EC2::Instance"), ("Custom::hoge::fuga")
])
def test_CfnSpecificationResourceType_valid(resource_type:str, context:AppContext):
    t = CfnSpecificationResourceTypeName(resource_type, context=context)
    assert t.fullname == resource_type

@pytest.mark.parametrize("resource_type", [
    ("AWS::EC2::Instance.BlockDeviceMappings"), ("AWS::EC2"), ("hogefuga")
])
def test_CfnSpecificationResourceType_invalid(resource_type:str, context:AppContext):
    with pytest.raises(AssertionError):
        CfnSpecificationResourceTypeName(resource_type, context=context)

@pytest.mark.parametrize("property_type,prefix,suffix", [
    ("AWS::EC2::Instance.BlockDeviceMapping", "AWS::EC2::Instance", "BlockDeviceMapping"),
    ("Custom::hoge::fuga.HOGEFUGA", "Custom::hoge::fuga", "HOGEFUGA"),
    ("Tag","Tag","Tag")
])
def test_CfnSpecificationPropertyType_valid(
    property_type:str, prefix:str, suffix:str,
    context:AppContext,
):
    t = CfnSpecificationPropertyTypeName(property_type, context=context)
    assert t.fullname == property_type
    assert t.prefix == prefix
    assert t.suffix == suffix

@pytest.mark.parametrize("property_type", [
    ("AWS::EC2::Instance"), ("BlockDeviceMapping"), ("AWS::EC2::Instance.BlockDeviceMapping.Ebs")
])
def test_CfnSpecificationPropertyType_invalid(property_type:str, context:AppContext):
    with pytest.raises(AssertionError):
        CfnSpecificationPropertyTypeName(property_type, context=context)



import logging
import os

import pytest
from cfn_docgen.adapters.internal.cache import LocalFileCache

from cfn_docgen.config import AppConfig, AppContext, AwsConnectionSettings, ConnectionSettings

@pytest.fixture
def context():
    return AppContext(
        log_level=logging.DEBUG, 
        connection_settings=ConnectionSettings(aws=AwsConnectionSettings(profile_name=None)),
    )

def test_LocalFileCache_put(context:AppContext):
    path1 = "/foo/bar.json"
    body1 = "foobar"
    cache = LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=context)

    cache.put(path1, body1)
    cache_filepath = os.path.join(
        cache.cache_root_dir, cache.hash(path1)
    )
    assert os.path.isfile(cache_filepath), f"{cache_filepath} does not exists"
    with open(cache_filepath, "r", encoding=cache.encoding) as fp:
        assert body1 == fp.read()


def test_LocalFileCache_get(context:AppContext):
    path1 = "/foo/bar.json"
    body1 = "foobar"
    cache = LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=context)
    
    cache.put(path1, body1)

    assert body1 == cache.get(path1)
    assert cache.get("not-exist") is None





import logging
import pytest
import os
import boto3

from cfn_docgen.entrypoints.serverless import lambda_function
from cfn_docgen.entrypoints.serverless.model.lambda_model import S3NotificationEvent, S3NotificationEventRecord, S3NotificationEventRecordS3, S3NotificationEventRecordS3Bucket, S3NotificationEventRecordS3Object

INPUT_MASTER_FILE=os.path.join(
    os.path.dirname(__file__),
    "..", "..", "..", "..", "..", "docs", "sample-template.yaml"
)
EXPECTED_MASTER_FILE=os.path.join(
    os.path.dirname(__file__),
    "..", "..", "..", "..", "..", "docs", "sample-template.md"
)

# s3 bucket and objects
s3_client = boto3.client("s3") # type: ignore
TEST_BUCKET_NAME=os.environ["TEST_BUCKET_NAME"]
INPUT_ROOT_PREFIX=f"s3://{TEST_BUCKET_NAME}/data/input"
INPUT_PREFIX1=INPUT_ROOT_PREFIX+"/dir1"
INPUT_KEY1=INPUT_PREFIX1+"/sample-template.yaml"
INPUT_PREFIX2=INPUT_PREFIX1+"/dir2"
INPUT_KEY2=INPUT_PREFIX2+"/sample-template.yaml"

OUTPUT_ROOT_PREFIX=f"s3://{TEST_BUCKET_NAME}/data/output"
OUTPUT_PREFIX1=OUTPUT_ROOT_PREFIX+"/dir1"
OUTPUT_MD_KEY1=OUTPUT_PREFIX1+"/sample-template.md"
OUTPUT_PREFIX2=OUTPUT_PREFIX1+"/dir2"
OUTPUT_MD_KEY2=OUTPUT_PREFIX2+"/sample-template.md"


@pytest.fixture(scope="class", autouse=True)
def class_s3_bucket_and_keys():
    # cleanup s3 keys
    for prefix in [INPUT_ROOT_PREFIX, OUTPUT_ROOT_PREFIX]:
        res = s3_client.list_objects(
            Bucket=TEST_BUCKET_NAME,
            Prefix=prefix,
        )
        contents = res.get("Contents", None)
        if contents is None or len(contents) == 0: # type: ignore
            continue
        s3_client.delete_objects(
            Bucket=TEST_BUCKET_NAME,
            Delete={
                "Objects": [
                    {"Key": content["Key"]} for content in contents # type: ignore
                ]
            }
        )
    
    # prepare input keys
    for key in [INPUT_KEY1, INPUT_KEY2]:
        s3_client.upload_file(
            Bucket=TEST_BUCKET_NAME,
            Key=key,
            Filename=INPUT_MASTER_FILE
        )


@pytest.fixture(scope="function", autouse=True)
def function_s3_bucket_and_keys():
    # cleanup s3 keys
    res = s3_client.list_objects_v2(
        Bucket=TEST_BUCKET_NAME,
        Prefix=OUTPUT_ROOT_PREFIX,
    )
    contents = res.get("Contents", None)
    if contents is not None and len(contents) > 0: # type: ignore
        s3_client.delete_objects(
            Bucket=TEST_BUCKET_NAME,
            Delete={
                "Objects": [
                    {"Key": content["Key"]} for content in contents # type: ignore
                ]
            }
        )


def test_serverless():
    event = S3NotificationEvent(
        Records=[
            S3NotificationEventRecord(
                s3=S3NotificationEventRecordS3(
                    bucket=S3NotificationEventRecordS3Bucket(
                        name=TEST_BUCKET_NAME
                    ),
                    object=S3NotificationEventRecordS3Object(
                        key=INPUT_KEY1.replace(f"s3://{TEST_BUCKET_NAME}/", "")
                    )
                )
            ),
            S3NotificationEventRecord(
                s3=S3NotificationEventRecordS3(
                    bucket=S3NotificationEventRecordS3Bucket(
                        name=TEST_BUCKET_NAME
                    ),
                    object=S3NotificationEventRecordS3Object(
                        key=INPUT_KEY2.replace(f"s3://{TEST_BUCKET_NAME}/", "")
                    )
                )
            )
        ]
    )
    result = lambda_function.lambda_handler(event=event.model_dump(), context=None)
    assert len(result) == 2

    with open(EXPECTED_MASTER_FILE, "rb") as fp:
        expected = fp.read()
    res = s3_client.get_object(
        Bucket=lambda_function.DEST_BUCKET_NAME, 
        Key=os.path.join(
            lambda_function.DEST_BUCKET_PREFIX, os.path.dirname(INPUT_KEY1.replace(f"s3://{TEST_BUCKET_NAME}/", "")), "sample-template.md"
        ),
    )
    assert res["Body"].read() == expected
    res = s3_client.get_object(
        Bucket=lambda_function.DEST_BUCKET_NAME, 
        Key=os.path.join(
            lambda_function.DEST_BUCKET_PREFIX, os.path.dirname(INPUT_KEY2.replace(f"s3://{TEST_BUCKET_NAME}/", "")), "sample-template.md"
        ),
    )
    assert res["Body"].read() == expected



def test_serverless_continue(caplog:pytest.LogCaptureFixture):
    caplog.set_level(logging.WARNING)
    try:
        event = S3NotificationEvent(
            Records=[
                S3NotificationEventRecord(
                    s3=S3NotificationEventRecordS3(
                        bucket=S3NotificationEventRecordS3Bucket(
                            name=TEST_BUCKET_NAME
                        ),
                        object=S3NotificationEventRecordS3Object(
                            key=INPUT_KEY1.replace(f"s3://{TEST_BUCKET_NAME}/", "")
                        )
                    )
                ),
                S3NotificationEventRecord(
                    s3=S3NotificationEventRecordS3(
                        bucket=S3NotificationEventRecordS3Bucket(
                            name=TEST_BUCKET_NAME
                        ),
                        object=S3NotificationEventRecordS3Object(
                            key=INPUT_KEY2.replace(f"s3://{TEST_BUCKET_NAME}/", "")
                        )
                    )
                )
            ]
        )
        # replace partial file with invalid body
        s3_client.put_object(Bucket=TEST_BUCKET_NAME, Key=INPUT_KEY1.replace(f"s3://{TEST_BUCKET_NAME}/", ""), Body=b"invalid")

        result = lambda_function.lambda_handler(event=event.model_dump(), context=None)
        assert len(result) == 1

        warn_messages = [r.message for r in caplog.records if r.levelno == logging.WARNING]

        assert "failed to generate document" in warn_messages[0]

    finally:
        with open(INPUT_MASTER_FILE, "rb") as fp:
            s3_client.put_object(Bucket=TEST_BUCKET_NAME, Key=INPUT_KEY1.replace(f"s3://{TEST_BUCKET_NAME}/", ""), Body=fp.read())




import json
from typing import List, Mapping
from cfn_docgen.config import AppContext

from cfn_docgen.domain.model.cfn_specification import CfnSpecificationForResource, CfnSpecificationPropertyTypeName, CfnSpecificationResourceTypeName, CfnSpecificationPropertyType, CfnSpecificationResourceType
from cfn_docgen.domain.model.cfn_specification import CfnSpecification
from cfn_docgen.domain.ports.cache import IFileCache
from cfn_docgen.domain.ports.internal.file_loader import IFileLoader
from cfn_docgen.domain.ports.cfn_specification_repository import ICfnSpecificationRepository


class CfnSpecificationRepository(ICfnSpecificationRepository):
    def __init__(
        self, 
        source_url:str,
        loader: IFileLoader, 
        cache: IFileCache, 
        recursive_resource_types:List[str],
        context:AppContext,
    ) -> None:
        self.context = context

        try:
            cached = cache.get(source_url)
            if cached is None:
                self.context.log_debug(f"cache not hit [{source_url}]")
                json_bytes = loader.download(source_url)
                json_str = json_bytes.decode()
                self.spec = CfnSpecification(**json.loads(json_str))
                cache.put(source_url, json_str)
            else:
                self.context.log_debug(f"cache hit [{source_url}]")
                self.spec = CfnSpecification(**json.loads(cached))
            self.recursive_resource_types = recursive_resource_types
        except Exception as ex:
            self.context.log_error("failed to setup CfnSpecificationRepository")
            raise ex


    def get_resource_spec(self, resource_type: CfnSpecificationResourceTypeName) -> CfnSpecificationResourceType:
        try:
            return self.spec.ResourceTypes[resource_type.fullname]
        except KeyError as ex:
            self.context.log_debug(f"Specification for resource tyoe [{resource_type.fullname}] is not found")
            raise ex

    def get_property_spec(self, property_type: CfnSpecificationPropertyTypeName) -> CfnSpecificationPropertyType:
        try:
            return self.spec.PropertyTypes[property_type.fullname]
        except KeyError as ex:
            self.context.log_debug(f"Specification for property tyoe [{property_type.fullname}] is not found")
            raise ex

    def list_properties_for_resource(self, resource_type: CfnSpecificationResourceTypeName) -> Mapping[str, CfnSpecificationPropertyType]:
        property_specs:Mapping[str, CfnSpecificationPropertyType] = {}
        for property_type, property_spec in self.spec.PropertyTypes.items():
            if property_type.startswith(f"{resource_type.fullname}.") or property_type == "Tag":
                property_specs[property_type] = property_spec
        return property_specs


    def get_specs_for_resource(self, resource_type: CfnSpecificationResourceTypeName) -> CfnSpecificationForResource:
        resource_spec = self.get_resource_spec(resource_type)
        property_specs = self.list_properties_for_resource(resource_type)
        return CfnSpecificationForResource(
            ResourceSpec=resource_spec,
            PropertySpecs=property_specs,
        )

    def is_recursive(self, resource_type: CfnSpecificationResourceTypeName) -> bool:
        if resource_type.fullname in self.recursive_resource_types:
            self.context.log_debug(f"resource type [{resource_type}] is recurisive resource")
            return True
        return False



from abc import ABC, abstractmethod
from typing import Optional

from cfn_docgen.config import AppContext

class IFileCache(ABC):

    def __init__(self, cache_root_dir:str, context:AppContext) -> None:
        self.cache_root_dir = cache_root_dir
        self.context = context

        super().__init__()

    @abstractmethod
    def put(self, filepath:str, body:str):
        pass
    @abstractmethod
    def get(self, filepath:str) -> Optional[str]:
        pass


from cfn_docgen.domain.model.cfn_document_generator import (
    CfnDocumentDestination as __CfnDocumentDestination
)
from cfn_docgen.domain.services.cfn_docgen_service import (
    CfnDocgenService as __CfnDocgenService,
    CfnDocgenServiceCommandInput as __CfnDocgenServiceCommandInput,
    CfnTemplateSource as __CfnTemplateSource
)


__version__="0.10.6"

CfnDocgenService = __CfnDocgenService
CfnDocgenServiceCommandInput = __CfnDocgenServiceCommandInput
CfnTemplateSource = __CfnTemplateSource
CfnDocumentDestination = __CfnDocumentDestination


import logging
import os
from typing import Any

import pytest
from cfn_docgen.adapters.cfn_document_storage import CfnDocumentStorage
from cfn_docgen.adapters.internal.file_loader import document_loader_factory
from cfn_docgen.config import AppContext, AwsConnectionSettings, ConnectionSettings
from cfn_docgen.domain.model.cfn_document_generator import CfnDocumentDestination

OUTPUT_FILE=os.path.join(
    os.path.dirname(__file__), "output.txt"
)

@pytest.fixture
def context():
    return AppContext(
        log_level=logging.DEBUG,
        connection_settings=ConnectionSettings(aws=AwsConnectionSettings(profile_name=None)),
    )

def teardown_function(function:Any):
    try:
        os.remove(OUTPUT_FILE)
    except FileNotFoundError:
        pass


def test_CfnDocumentStorage_save_document(context:AppContext):

    storage = CfnDocumentStorage(
        document_loader_factory=document_loader_factory,
        context=context,
    )
    body = b"body"
    dest = CfnDocumentDestination(
        dest=OUTPUT_FILE,
        context=context,
    )
    storage.save_document(body, dest)

    with open(OUTPUT_FILE, "rb") as fp:
        assert fp.read() == body




import logging
import os
from typing import Any

import boto3
import pytest
from cfn_docgen.adapters.cfn_document_storage import document_storage_facotory
from cfn_docgen.adapters.cfn_specification_repository import CfnSpecificationRepository
from cfn_docgen.adapters.cfn_template_provider import template_provider_factory
from cfn_docgen.adapters.internal.cache import LocalFileCache
from cfn_docgen.adapters.internal.file_loader import RemoteFileLoader
from cfn_docgen.config import AppConfig, AppContext, AwsConnectionSettings, ConnectionSettings
from cfn_docgen.domain.model.cfn_document_generator import CfnDocumentDestination, document_generator_factory
from cfn_docgen.domain.model.cfn_template import CfnTemplateSource
from cfn_docgen.domain.services.cfn_docgen_service import CfnDocgenService, CfnDocgenServiceCommandInput

INPUT_FILE=os.path.join(
    os.path.dirname(__file__), "..", "..", "..", "..", "..", "docs", "sample-template.yaml",
)
OUTPUT_FILE=os.path.abspath(os.path.join(".", "output.md"))
EXAMPLE_FILE=os.path.join(
    os.path.dirname(__file__), "..", "..", "..", "..", "..", "docs", "sample-template.md",
)
TEST_BUCKET_NAME=os.environ["TEST_BUCKET_NAME"]
TEST_INPUT_KEY="unit-test/sample-template.yaml"
TEST_OUTPUT_KEY="unit-test/sample-template.md"

s3 = boto3.client("s3") # type: ignore
def setup_module(module:Any):
    for key in [TEST_INPUT_KEY, TEST_OUTPUT_KEY]:
        try:
            s3.delete_object(
                Bucket=TEST_BUCKET_NAME,
                Key=key
            )
        except Exception:
            pass

def teardown_function(function:Any):
    try:
        os.remove(OUTPUT_FILE)
    except FileNotFoundError:
        pass

@pytest.fixture
def context():
    return AppContext(
        log_level=logging.DEBUG,
        connection_settings=ConnectionSettings(aws=AwsConnectionSettings(profile_name=None)),
    )

def test_CfnDocgenService_main_localfile(context:AppContext):
    service = CfnDocgenService(
        context=context,
        cfn_template_provider_facotry=template_provider_factory,
        cfn_document_generator_factory=document_generator_factory,
        cfn_document_storage_factory=document_storage_facotory,
        cfn_specification_repository=CfnSpecificationRepository(
            context=context,
            source_url=AppConfig.DEFAULT_SPECIFICATION_URL,
            loader=RemoteFileLoader(context=context),
            cache=LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=context),
            recursive_resource_types=AppConfig.RECURSIVE_RESOURCE_TYPES,
        )

    )
    command_input = CfnDocgenServiceCommandInput(
        template_source=CfnTemplateSource(source=INPUT_FILE, context=context),
        document_dest=CfnDocumentDestination(OUTPUT_FILE, context=context),
        fmt="markdown",
    )
    command_output = service.main(command_input=command_input)
    
    assert command_output.document_dest == OUTPUT_FILE

    with open(OUTPUT_FILE, "r") as ofp:
        with open(EXAMPLE_FILE, "r") as efp:
            output = ofp.read()
            expected = efp.read()

    assert output == expected

    
def test_CfnDocgenService_main_s3obj(context:AppContext):
    s3.upload_file(
        Bucket=TEST_BUCKET_NAME,
        Key=TEST_INPUT_KEY,
        Filename=INPUT_FILE,
    )
    service = CfnDocgenService(
        context=context,
        cfn_template_provider_facotry=template_provider_factory,
        cfn_document_generator_factory=document_generator_factory,
        cfn_document_storage_factory=document_storage_facotory,
        cfn_specification_repository=CfnSpecificationRepository(
            context=context,
            source_url=AppConfig.DEFAULT_SPECIFICATION_URL,
            loader=RemoteFileLoader(context=context),
            cache=LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=context),
            recursive_resource_types=AppConfig.RECURSIVE_RESOURCE_TYPES,
        )

    )
    command_input = CfnDocgenServiceCommandInput(
        template_source=CfnTemplateSource(
            source=f"s3://{TEST_BUCKET_NAME}/{TEST_INPUT_KEY}",
            context=context,
        ),
        document_dest=CfnDocumentDestination(
            dest=f"s3://{TEST_BUCKET_NAME}/{TEST_OUTPUT_KEY}",
            context=context,
        ),
        fmt="markdown",
    )
    command_output = service.main(command_input=command_input)
    
    assert command_output.document_dest == f"s3://{TEST_BUCKET_NAME}/{TEST_OUTPUT_KEY}"

    res = s3.get_object(Bucket=TEST_BUCKET_NAME, Key=TEST_OUTPUT_KEY)
    with open(EXAMPLE_FILE, "rb") as fp:
        assert res["Body"].read() == fp.read()

    



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
import datetime
import json
import re
from typing import Any, List, Literal, Mapping, cast
from cfn_docgen.config import AppContext
from cfn_docgen.domain.model.cfn_specification import CfnSpecificationProperty

from cfn_docgen.domain.model.cfn_template import CfnTemplateParameterDefinition, CfnTemplateResourcePropertiesNode, CfnTemplateResourcePropertyNode, CfnTemplateTree

SupportedFormat = Literal["markdown"]


def document_generator_factory(fmt:SupportedFormat, context:AppContext) -> ICfnDocumentGenerator:
    match fmt:
        case "markdown":
            context.log_debug(f"format is [{fmt}]. return {CfnMarkdownDocumentGenerator.__name__}")
            return CfnMarkdownDocumentGenerator(context=context)
        case _: # type: ignore
            raise NotImplementedError

class CfnDocumentDestination:
    type: Literal["LocalFilePath", "S3BucketKey", "HttpUrl"]
    dest: str

    def __init__(self, dest:str, context:AppContext) -> None:
        if dest.startswith("s3://"):
            self.type = "S3BucketKey"
        elif dest.startswith("http://") or dest.startswith("https://"):
            self.type = "HttpUrl"
        else:
            self.type = "LocalFilePath"
        self.dest = dest

        context.log_debug(f"type of dest [{dest}] is [{self.type}]")


@dataclass
class PropertyField:
    Property: str
    Value: str
    Description: str
    Required: str
    UpdateType: str
    Type: str
    def as_table_row(self) -> str:
        return "|{Property}|{Value}|{Description}|{Type}|{Required}|{UpdateType}|".format(
            **self.__dict__
        )

class ICfnDocumentGenerator(ABC):

    def __init__(self, context:AppContext) -> None:
        self.context = context

    @abstractmethod
    def generate(self, cfn_template_tree:CfnTemplateTree) -> str:
        pass

class CfnMarkdownDocumentGenerator(ICfnDocumentGenerator):
    lfc = "\n"
    failed_message = "**[cfn-docgen] FAILED TO GENERATE {section}**"

    def generate(self, cfn_template_tree: CfnTemplateTree) -> str:
        start = datetime.datetime.now()
        doc = self.lfc.join([
            self.header(),
            self.table_of_contents(cfn_template_tree),
            "",
            "---",
            "",
            self.overview(cfn_template_tree),
            "",
            "---",
            "",
            self.description(cfn_template_tree),
            "",
            "---",
            "",
            self.parameters(cfn_template_tree),
            "",
            "---",
            "",
            self.mappings(cfn_template_tree),
            "",
            "---",
            "",
            self.conditions(cfn_template_tree),
            "",
            "---",
            "",
            self.rules(cfn_template_tree),
            "",
            "---",
            "",
            self.resources(cfn_template_tree),
            "",
            "---",
            "",
            self.outputs(cfn_template_tree),
        ])

        end = datetime.datetime.now()
        self.context.log_debug(f"elapsed time to generate document for [{cfn_template_tree.template_source.source}] is [{(end - start).total_seconds()}s]")
        return doc

    def header(self) -> str:
        from cfn_docgen import __version__
        return f"<!-- cfn-docgen: generated by v{__version__} -->"


    def _toc_escape(self, s:str) -> str:
        """
        - replace ' ' with '-'
        - truncate symbols(except '_' and '-')
        - convert Upper to lower
        """
        return re.sub(
            r"[^\w_-]", "",
            s.lower().replace(" ", "-"),
        )

    def table_of_contents(self, tree:CfnTemplateTree) -> str:
        try:
            toc:List[str] = []
            def line(name:str, indent:int) -> str:
                return f"{' '*indent}- [{name}](#{self._toc_escape(name)})"

            toc.append(line(name=tree.template_source.basename, indent=0))
            toc.append(line(name="Description", indent=2))
            
            toc.append(line(name="Parameters", indent=2))
            for group_name in sorted(
                tree.parameters_node.group_nodes.keys()
            ):
                # list parameters in independ node at the tail of list
                if group_name == tree.parameters_node.group_name_for_independent_parameters:
                    continue
                toc.append(line(name=group_name, indent=4))
                for param_name in sorted(
                    tree.parameters_node.group_nodes[group_name].leaves.keys()
                ):
                    toc.append(line(name=param_name, indent=6))
            independent_nodes = tree.parameters_node.group_nodes.get(
                tree.parameters_node.group_name_for_independent_parameters,
            )
            if independent_nodes is not None:
                for param_name in sorted(independent_nodes.leaves.keys()):
                    toc.append(line(name=param_name, indent=4))
            
            toc.append(line(name="Mappings", indent=2))
            for map_name in sorted(tree.mappings_node.mapping_leaves.keys()):
                toc.append(line(name=map_name, indent=4))
            
            toc.append(line(name="Conditions", indent=2))
            for cond_name in sorted(tree.conditions_node.condition_leaves.keys()):
                toc.append(line(name=cond_name, indent=4))
            
            toc.append(line(name="Rules", indent=2))
            for rule_name in sorted(tree.rules_node.rule_leaves.keys()):
                toc.append(line(name=rule_name, indent=4))
            
            toc.append(line(name="Resources", indent=2))
            for group_name in sorted(
                tree.resources_node.group_nodes.keys(),
            ):
                if group_name == tree.resources_node.group_name_for_independent_resources:
                    continue
                toc.append(line(name=group_name, indent=4))
                for resource_name, resource_node in sorted(
                    tree.resources_node.group_nodes[group_name].resource_nodes.items(),
                    key=lambda item: (item[1].type, item[0])
                ):
                    toc.append(line(name=f"{resource_name} ({resource_node.type})", indent=6))
            independent_nodes = tree.resources_node.group_nodes.get(
                tree.resources_node.group_name_for_independent_resources,
            )
            if independent_nodes is not None:
                for resource_name, resource_node in sorted(
                    independent_nodes.resource_nodes.items(),
                    key=lambda item: (item[1].type, item[0])
                ):
                    toc.append(line(name=f"{resource_name} ({resource_node.type})", indent=4))
            
            toc.append(line(name="Outputs", indent=2))
            for out_name in sorted(tree.outputs_node.output_leaves.keys()):
                toc.append(line(name=out_name, indent=4))

            return self.lfc.join(toc)
        except Exception:
            self.context.log_warning(
                f"failed to build table_of_contents for [{tree.template_source.source}]"
            )
            return self.failed_message.format(section="Table of Contents")


    def _unordered_list(self, l:Any) -> str:
        if not isinstance(l, list):
            return str(l)
        
        l = cast(List[Any], l)
        if len(l) == 0:
            return "-"
        unorderd_list =  "<ul>"
        for elem in l:
            unorderd_list += f"<li>{str(elem)}</li>"
        unorderd_list += "</ul>"
        return unorderd_list

    def overview(self, tree:CfnTemplateTree) -> str:
        try:
            overview:List[str] = []

            version = "-"
            description = "-"
            transform = "-"
            if tree.aws_template_format_version is not None:
                version = tree.aws_template_format_version
            if tree.description is not None:
                description = tree.description
            if isinstance(tree.transform, str):
                transform = tree.transform
            else:
                transform = self._unordered_list(tree.transform)

            overview.append(f"# {tree.template_source.basename}")
            overview.append("")
            overview.append("| | |")
            overview.append("|-|-|")
            overview.append(f"|AWSTemplateFormatVersion|{version}|")
            overview.append(f"|Description|{description}|")
            overview.append(f"|Transform|{transform}|")

            return self.lfc.join(overview)
        except Exception:
            self.context.log_warning(
                f"failed to build overview for [{tree.template_source.source}]"
            )
            return self.failed_message.format(section="Overview")
    
    def description(self, tree:CfnTemplateTree) -> str:
        try:
            description:List[str] = []

            cfn_docgen_description = ""
            if tree.cfn_docgen_description is not None:
                cfn_docgen_description = tree.cfn_docgen_description
            
            description.append("## Description")
            description.append("")
            description.append(cfn_docgen_description)

            if description[-1] == "":
                description = description[:-1]

            return self.lfc.join(description)
        except Exception:
            self.context.log_warning(
                f"failed to build description for [{tree.template_source.source}]"
            )
            return self.failed_message.format(section="Description")
    
    def parameters(self, tree:CfnTemplateTree) -> str:
        try:
            parameters:List[str] = []
            headers = "|Type|Default|AllowedValues|AllowedPattern|NoEcho|MinValue|MaxValue|MinLength|MaxLength|ConstraintDescription|"
            separators = "|-|-|-|-|-|-|-|-|-|-|"

            def row(d:CfnTemplateParameterDefinition) -> str:
                return "|{Type}|{Default}|{AllowedValues}|{AllowedPattern}|{NoEcho}|{MinValue}|{MaxValue}|{MinLength}|{MaxLength}|{ConstraintDescription}|".format(
                    Type=self._dump_json(d.Type),
                    Default=self._unordered_list(d.Default) if d.Default is not None else "-",
                    AllowedValues=self._unordered_list(d.AllowedValues) if d.AllowedValues is not None else "-",
                    AllowedPattern=d.AllowedPattern if d.AllowedPattern is not None else "-",
                    NoEcho=str(d.NoEcho).lower(),
                    MinValue=str(d.MinValue) if d.MinValue is not None else "-",
                    MaxValue=str(d.MaxValue) if d.MaxValue is not None else "-",
                    MinLength=str(d.MinLength) if d.MinLength is not None else "-",
                    MaxLength=str(d.MaxLength) if d.MaxLength is not None else "-",
                    ConstraintDescription=d.ConstraintDescription if d.ConstraintDescription is not None else "-",
                )

            parameters.append("## Parameters")
            parameters.append("")

            for group_name, group_node in sorted(tree.parameters_node.group_nodes.items(), key=lambda g: g[0]):
                if group_name == tree.parameters_node.group_name_for_independent_parameters:
                    continue
                parameters.append(f"### {group_name}")
                parameters.append("")
                for param_name, param in sorted(group_node.leaves.items(), key=lambda p: p[0]):
                    try:
                        d = param.definition
                        parameters.append(f"#### {param_name}")
                        parameters.append("")
                        parameters.append(
                            d.Description if d.Description is not None else ""
                        )
                        parameters.append("")
                        parameters.append(headers)
                        parameters.append(separators)
                        parameters.append(row(param.definition))
                        parameters.append("")
                    except Exception:
                        self.context.log_warning(f"failed to build parameter [{param_name}]")
                        continue

            independent_node = tree.parameters_node.group_nodes.get(
                tree.parameters_node.group_name_for_independent_parameters,
            )
            if independent_node is not None:
                for param_name, param in sorted(
                    independent_node.leaves.items(), key=lambda p: p[0],
                ):
                    try:
                        d = param.definition
                        parameters.append(f"### {param_name}")
                        parameters.append("")
                        parameters.append(
                            d.Description if d.Description is not None else ""
                        )
                        parameters.append("")
                        parameters.append(headers)
                        parameters.append(separators)
                        parameters.append(row(param.definition))
                        parameters.append("")
                    except Exception:
                        self.context.log_warning(f"failed to build parameter [{param_name}]")
                        continue

            if parameters[-1] == "":
                parameters = parameters[:-1]

            return self.lfc.join(parameters)
        except Exception:
            self.context.log_warning(
                f"failed to build parameters for [{tree.template_source.source}]"
            )
            return self.failed_message.format(section="Parameters")
    
    def mappings(self, tree:CfnTemplateTree) -> str:
        try:
            mappings:List[str] = []
            
            mappings.append("## Mappings")
            mappings.append("")
            for map_name, map_leaf in sorted(tree.mappings_node.mapping_leaves.items(), key=lambda m: m[0]):
                try:
                    mappings.append(f"### {map_name}")
                    mappings.append("")
                    mappings.append(map_leaf.description if map_leaf.description is not None else "")
                    mappings.append("")
                    mappings.append("|Map|Key|Value|")
                    mappings.append("|-|-|-|")
                    for name, body in sorted(map_leaf.definition.items(), key=lambda i: i[0]):
                        for key, val in sorted(body.items(), key=lambda _i: _i[0]):
                            mappings.append(f"|{name}|{key}|{str(val)}|")
                    mappings.append("")
                except Exception:
                    self.context.log_warning(f"failed to build mapping [{map_name}]")
                    continue

            if mappings[-1] == "":
                mappings = mappings[:-1]

            return self.lfc.join(mappings)
        except Exception:
            self.context.log_warning(
                f"failed to build mappings for [{tree.template_source.source}]"
            )
            return self.failed_message.format(section="Mappings")
        
    def _dump_json(self, j:Any) -> str:
        if not isinstance(j, dict) and not isinstance(j, list):
            if isinstance(j, bool):
                return str(j).lower()
            
            return str(j).replace(
                "$", "\\$"
            ).replace(
                "|", "\\|"
            ).replace(
                "<", "\\<",
            ).replace(
                ">", "\\>"
            )
        
        dumped = json.dumps(j, indent=2, ensure_ascii=False).replace(
            "$", "\\$"
        ).replace(
            "|", "\\|"
        ).replace(
            "<", "\\<",
        ).replace(
            ">", "\\>"
        )
        dumped = dumped.replace("\n", "<br/>").replace(" ", "&nbsp;")
        return dumped

        

    def conditions(self, tree:CfnTemplateTree) -> str:
        try:
            conditions:List[str] = []

            conditions.append("## Conditions")
            conditions.append("")
            for cond_name, cond_leaf in sorted(tree.conditions_node.condition_leaves.items(), key=lambda c: c[0]):
                try:
                    conditions.append(f"### {cond_name}")
                    conditions.append("")
                    conditions.append(cond_leaf.descirption if cond_leaf.descirption is not None else "")
                    conditions.append("")
                    conditions.append("|Condition|")
                    conditions.append("|-|")
                    conditions.append(f"|{self._dump_json(cond_leaf.definition)}|")
                    conditions.append("")
                except Exception:
                    self.context.log_warning(f"failed to build condition [{cond_name}]")

            if conditions[-1] == "":
                conditions = conditions[:-1]

            return self.lfc.join(conditions)
        except Exception:
            self.context.log_warning(
                f"failed to build conditions for [{tree.template_source.source}]"
            )
            return self.failed_message.format(section="Conditions")
    

    def rules(self, tree:CfnTemplateTree) -> str:
        try:
            rules:List[str] = []

            rules.append("## Rules")
            rules.append("")
            for rule_name, rule_leaf in sorted(tree.rules_node.rule_leaves.items(), key=lambda r: r[0]):
                try:
                    rules.append(f"### {rule_name}")
                    rules.append("")
                    rules.append(rule_leaf.description if rule_leaf.description is not None else "")
                    rules.append("")
                    rules.append("|RuleCondition|")
                    rules.append("|-|")
                    rules.append("|{RuleCondition}|".format(
                        RuleCondition=self._dump_json(
                            rule_leaf.definition.RuleCondition
                        ) if rule_leaf.definition.RuleCondition is not None else '-'
                    ))
                    rules.append("")
                    rules.append("|Assert|AssertDescription|")
                    rules.append("|-|-|")
                    for assertion in rule_leaf.definition.Assertions:
                        rules.append("|{Assert}|{AssertDescription}|".format(
                            Assert=self._dump_json(assertion.Assert),
                            AssertDescription=self._dump_json(assertion.AssertDescription)
                        ))
                    rules.append("")
                except Exception:
                    self.context.log_warning(f"failed to build rule [{rule_name}]")
                    continue

            if rules[-1] == "":
                rules = rules[:-1]

            return self.lfc.join(rules)
        except Exception:
            self.context.log_warning(
                f"failed to build rules for [{tree.template_source.source}]"
            )
            return self.failed_message.format(section="Rules")

    def _prop_type_rep(self, p:CfnSpecificationProperty) -> str:
        if p.PrimitiveType is not None:
            return p.PrimitiveType
        if p.PrimitiveItemType is not None and p.Type is not None:
            return f"{p.Type} of {p.PrimitiveItemType}"
        if p.ItemType is not None and p.Type is not None:
            return f"{p.Type} of {p.ItemType}"
        if p.Type is not None:
            return p.Type
        
        return "-"

    def _simplify_jsonpath(self, json_path:str) -> str:
        *rest, tail = json_path.split(".")
        if len(rest) == 1: # ["$"]
            return tail
        
        spaces = "&nbsp;" * 2 * (len(rest)-1)
        # index = re.search(r"\[\d\]$", rest[-1])
        # if index is not None:
        #     return f"{spaces}{index.group()}.{tail}"
        
        return f"{spaces}{tail}"

    def _flatten_property_node(self, property_node:CfnTemplateResourcePropertyNode) -> Mapping[str, PropertyField]:
        property_fields:Mapping[str, PropertyField] = {}

        for leaf in property_node.property_leaves.values():
            if leaf.definition is None:
                continue
            property_fields[leaf.json_path] = PropertyField(
                Property=self._simplify_jsonpath(leaf.json_path),
                Value=self._dump_json(leaf.definition),
                Description=leaf.description if leaf.description is not None else "-",
                Required=str(leaf.spec.Required).lower() if leaf.spec.Required is not None else "-",
                UpdateType=leaf.spec.UpdateType if leaf.spec.UpdateType is not None else "-",
                Type=self._prop_type_rep(leaf.spec)
            )
        for node in property_node.property_nodes.values():
            if not node.has_leaves:
                continue
            property_fields[node.json_path] = PropertyField(
                Property=self._simplify_jsonpath(node.json_path),
                Value="-",
                Description=node.description if node.description is not None else "-",
                Required=str(node.spec.Required).lower() if node.spec is not None and node.spec.Required is not None else "-",
                UpdateType=node.spec.UpdateType if node.spec is not None and node.spec.UpdateType is not None else "-",
                Type=self._prop_type_rep(node.spec) if node.spec is not None else "-"
            )
            property_fields.update(self._flatten_property_node(node))
        for nodes in property_node.property_nodes_list.values():
            for node in nodes:
                if not node.has_leaves:
                    continue
                property_fields[node.json_path] = PropertyField(
                    Property=self._simplify_jsonpath(node.json_path),
                    Value="-",
                    Description=node.description if node.description is not None else "-",
                    Required=str(node.spec.Required).lower() if node.spec is not None and node.spec.Required is not None else "-",
                    UpdateType=node.spec.UpdateType if node.spec is not None and node.spec.UpdateType is not None else "-",
                    Type=self._prop_type_rep(node.spec) if node.spec is not None else "-"
                )
                property_fields.update(self._flatten_property_node(node))
        for nodes in property_node.property_nodes_map.values():
            for node in nodes.values():
                if not node.has_leaves:
                    continue
                property_fields[node.json_path] = PropertyField(
                    Property=self._simplify_jsonpath(node.json_path),
                    Value="-",
                    Description=node.description if node.description is not None else "-",
                    Required=str(node.spec.Required).lower() if node.spec is not None and node.spec.Required is not None else "-",
                    UpdateType=node.spec.UpdateType if node.spec is not None and node.spec.UpdateType is not None else "-",
                    Type=self._prop_type_rep(node.spec) if node.spec is not None else "-"
                )
                property_fields.update(self._flatten_property_node(node))

        return property_fields

    def _flatten_properties_node(self, properties_node:CfnTemplateResourcePropertiesNode) -> List[PropertyField]:
        property_fields:Mapping[str, PropertyField] = {}

        # leaves
        for leaf in properties_node.property_leaves.values():
            if leaf.definition is None:
                continue
            property_fields[leaf.json_path] = PropertyField(
                Property=self._simplify_jsonpath(leaf.json_path),
                Value=self._dump_json(leaf.definition),
                Description=leaf.description if leaf.description is not None else "-",
                Required=str(leaf.spec.Required).lower() if leaf.spec.Required is not None else "-",
                UpdateType=leaf.spec.UpdateType if leaf.spec.UpdateType is not None else "-",
                Type=self._prop_type_rep(leaf.spec)
            )
        for node in properties_node.property_nodes.values():
            if not node.has_leaves:
                continue
            property_fields[node.json_path] = PropertyField(
                Property=self._simplify_jsonpath(node.json_path),
                Value="-",
                Description=node.description if node.description is not None else "-",
                Required=str(node.spec.Required).lower() if node.spec is not None and node.spec.Required is not None else "-",
                UpdateType=node.spec.UpdateType if node.spec is not None and node.spec.UpdateType is not None else "-",
                Type=self._prop_type_rep(node.spec) if node.spec is not None else "-"
            )
            property_fields.update(self._flatten_property_node(node))
        for nodes in properties_node.property_nodes_list.values():
            for node in nodes:
                if not node.has_leaves:
                    continue
                property_fields[node.json_path] = PropertyField(
                    Property=self._simplify_jsonpath(node.json_path),
                    Value="-",
                    Description=node.description if node.description is not None else "-",
                    Required=str(node.spec.Required).lower() if node.spec is not None and node.spec.Required is not None else "-",
                    UpdateType=node.spec.UpdateType if node.spec is not None and node.spec.UpdateType is not None else "-",
                    Type=self._prop_type_rep(node.spec) if node.spec is not None else "-"
                )
                property_fields.update(self._flatten_property_node(node))
        for nodes in properties_node.property_nodes_map.values():
            for node in nodes.values():
                if not node.has_leaves:
                    continue
                property_fields[node.json_path] = PropertyField(
                    Property=self._simplify_jsonpath(node.json_path),
                    Value="-",
                    Description=node.description if node.description is not None else "-",
                    Required=str(node.spec.Required).lower() if node.spec is not None and node.spec.Required is not None else "-",
                    UpdateType=node.spec.UpdateType if node.spec is not None and node.spec.UpdateType is not None else "-",
                    Type=self._prop_type_rep(node.spec) if node.spec is not None else "-"
                )
                property_fields.update(self._flatten_property_node(node))

        property_fields_tuple = sorted(
            [(k, v) for k, v in property_fields.items()],
            key=lambda t: t[0]
        )
        property_fields_list = [t[1] for t in property_fields_tuple]
        return property_fields_list

    def resources(self, tree:CfnTemplateTree) -> str:
        try:
            resources:List[str] = []


            resources.append("## Resources")
            resources.append("")
            # for resource_name, resource_node in sorted(
            #     tree.resources_node.resource_nodes.items(), key=lambda r: (r[1].type, r[0])
            # ):
            #     try:
                    
            #         resources.append("### [{Name} ({Type})]({Url})".format(
            #             Name=resource_name,
            #             Type=resource_node.type,
            #             Url=resource_node.spec.Documentation,
            #         ))
            #         resources.append("")
            #         resources.append(
            #             resource_node.description if resource_node.description is not None else ""
            #         )
            #         resources.append("")
            #         resources.append("|DependsOn|Condition|CreationPolicy|UpdatePolicy|UpdateReplacePolicy|DeletionPolicy|")
            #         resources.append("|-|-|-|-|-|-|")
            #         resources.append("|{DependsOn}|{Condition}|{CreationPolicy}|{UpdatePolicy}|{UpdateReplacePolicy}|{DeletionPolicy}|".format(
            #             DependsOn=self._dump_json(resource_node.depends_on) if len(resource_node.depends_on) > 0 else "-",
            #             Condition=resource_node.condition if resource_node.condition is not None else "-",
            #             CreationPolicy=self._dump_json(resource_node.creation_policy) if resource_node.creation_policy is not None else "-",
            #             UpdatePolicy=self._dump_json(resource_node.update_policy) if resource_node.update_policy is not None else "-",
            #             UpdateReplacePolicy=resource_node.update_replace_policy,
            #             DeletionPolicy=resource_node.deletion_policy,
            #         ))
            #         resources.append("")

            #         resources.append("|Property|Value|Description|Type|Required|UpdateType|")
            #         resources.append("|-|-|-|-|-|-|")
            #         property_fields = self._flatten_properties_node(resource_node.properties_node)
            #         for property_field in property_fields:
            #             resources.append(property_field.as_table_row())
            #         resources.append("")
            #     except Exception:
            #         self.context.log_warning(f"failed to build resource [{resource_name}]")

            for group_name, group_node in sorted(
                tree.resources_node.group_nodes.items(), key=lambda item: item[0]
            ):
                try:
                    if group_name == tree.resources_node.group_name_for_independent_resources:
                        continue
                    resources.append(f"### {group_name}")
                    resources.append("")
                    for resource_name, resource_node in sorted(
                        group_node.resource_nodes.items(), key= lambda item: (item[1].type, item[0])
                    ):
                        try:
                            
                            resources.append("#### [{Name} ({Type})]({Url})".format(
                                Name=resource_name,
                                Type=resource_node.type,
                                Url=resource_node.spec.Documentation,
                            ))
                            resources.append("")
                            resources.append(
                                resource_node.description if resource_node.description is not None else ""
                            )
                            resources.append("")
                            resources.append("|DependsOn|Condition|CreationPolicy|UpdatePolicy|UpdateReplacePolicy|DeletionPolicy|")
                            resources.append("|-|-|-|-|-|-|")
                            resources.append("|{DependsOn}|{Condition}|{CreationPolicy}|{UpdatePolicy}|{UpdateReplacePolicy}|{DeletionPolicy}|".format(
                                DependsOn=self._dump_json(resource_node.depends_on) if len(resource_node.depends_on) > 0 else "-",
                                Condition=resource_node.condition if resource_node.condition is not None else "-",
                                CreationPolicy=self._dump_json(resource_node.creation_policy) if resource_node.creation_policy is not None else "-",
                                UpdatePolicy=self._dump_json(resource_node.update_policy) if resource_node.update_policy is not None else "-",
                                UpdateReplacePolicy=resource_node.update_replace_policy,
                                DeletionPolicy=resource_node.deletion_policy,
                            ))
                            resources.append("")

                            resources.append("|Property|Value|Description|Type|Required|UpdateType|")
                            resources.append("|-|-|-|-|-|-|")
                            property_fields = self._flatten_properties_node(resource_node.properties_node)
                            for property_field in property_fields:
                                resources.append(property_field.as_table_row())
                            resources.append("")
                        except Exception:
                            self.context.log_warning(f"failed to build resource [{resource_name}]")
                except Exception:
                    self.context.log_warning(f"failed to build resource group [{group_name}]")

            independent_node = tree.resources_node.group_nodes.get(
                tree.resources_node.group_name_for_independent_resources,
            )
            if independent_node is not None:
                for resource_name, resource_node in sorted(
                    independent_node.resource_nodes.items(), key= lambda item: (item[1].type, item[0])
                ):
                    try:
                        resources.append("### [{Name} ({Type})]({Url})".format(
                            Name=resource_name,
                            Type=resource_node.type,
                            Url=resource_node.spec.Documentation,
                        ))
                        resources.append("")
                        resources.append(
                            resource_node.description if resource_node.description is not None else ""
                        )
                        resources.append("")
                        resources.append("|DependsOn|Condition|CreationPolicy|UpdatePolicy|UpdateReplacePolicy|DeletionPolicy|")
                        resources.append("|-|-|-|-|-|-|")
                        resources.append("|{DependsOn}|{Condition}|{CreationPolicy}|{UpdatePolicy}|{UpdateReplacePolicy}|{DeletionPolicy}|".format(
                            DependsOn=self._dump_json(resource_node.depends_on) if len(resource_node.depends_on) > 0 else "-",
                            Condition=resource_node.condition if resource_node.condition is not None else "-",
                            CreationPolicy=self._dump_json(resource_node.creation_policy) if resource_node.creation_policy is not None else "-",
                            UpdatePolicy=self._dump_json(resource_node.update_policy) if resource_node.update_policy is not None else "-",
                            UpdateReplacePolicy=resource_node.update_replace_policy,
                            DeletionPolicy=resource_node.deletion_policy,
                        ))
                        resources.append("")

                        resources.append("|Property|Value|Description|Type|Required|UpdateType|")
                        resources.append("|-|-|-|-|-|-|")
                        property_fields = self._flatten_properties_node(resource_node.properties_node)
                        for property_field in property_fields:
                            resources.append(property_field.as_table_row())
                        resources.append("")
                    except Exception:
                        self.context.log_warning(f"failed to build resource [{resource_name}]")

            if resources[-1] == "":
                resources = resources[:-1]

            return self.lfc.join(resources)
        except Exception:
            self.context.log_warning(
                f"failed to build resources for [{tree.template_source.source}]"
            )
            return self.failed_message.format(section="Resources")
    
    def outputs(self, tree:CfnTemplateTree) -> str:
        try:
            outputs:List[str] = []

            outputs.append("## Outputs")
            outputs.append("")
            for output_name, output_leaf in sorted(
                tree.outputs_node.output_leaves.items(), key=lambda o: o[0],
            ):
                try:
                    outputs.append(f"### {output_name}")
                    outputs.append("")
                    outputs.append(
                        output_leaf.definition.Description if output_leaf.definition.Description is not None else ""
                    )
                    outputs.append("")
                    outputs.append("|Value|ExportName|Condition|")
                    outputs.append("|-|-|-|")
                    outputs.append("|{Value}|{ExportName}|{Condition}|".format(
                        Value=self._dump_json(output_leaf.definition.Value),
                        ExportName=self._dump_json(output_leaf.definition.Export.Name) if output_leaf.definition.Export is not None else  "-",
                        Condition=output_leaf.definition.Condition if output_leaf.definition.Condition is not None else "-",
                    ))
                    outputs.append("")
                except Exception:
                    self.context.log_warning(f"failed to build output [{output_name}]")
            
            if outputs[-1] == "":
                outputs = outputs[:-1]

            return self.lfc.join(outputs)

        except Exception:
            self.context.log_warning(
                f"failed to build outputs for [{tree.template_source.source}]"
            )
            return self.failed_message.format(section="Outputs")
        



from abc import ABC, abstractmethod
from typing import Callable
from cfn_docgen.config import AppContext
from cfn_docgen.domain.model.cfn_document_generator import CfnDocumentDestination

from cfn_docgen.domain.ports.internal.file_loader import IFileLoader


class ICfnDocumentStorage(ABC):

    def __init__(
        self, 
        document_loader_factory:Callable[[CfnDocumentDestination, AppContext], IFileLoader],
        context:AppContext,
    ) -> None:
        self.context = context
        self.document_loader_factory = document_loader_factory

    @abstractmethod
    def save_document(self, body:bytes, document_dest:CfnDocumentDestination) -> None:
        pass